import lxml.etree as x_xml
import numpy as np
import collections
__author__ = 'Jaime Undurraga'


def xml_file_to_dict(file_name=''):
    fxml = x_xml.parse(file_name)
    xml_root = fxml.getroot()
    return get_nodes(xml_root)


def xml_string_to_dict(string_content=''):
    xml_root = x_xml.fromstring(string_content)
    return get_nodes(xml_root)


def get_nodes(in_node=[]):
    c = in_node.getchildren()
    out = {}
    if len(c) > 0:
        # check that names are not repeated
        _children_tags = []
        for i in range(len(c)):
            _children_tags.append(c[i].tag)
        repeated = [x for x, y in list(collections.Counter(_children_tags).items()) if y > 1]

        list_out = []
        for i in range(len(c)):
            _node = get_nodes(c[i])
            if c[i].tag in repeated:
                list_out.append(_node[c[i].tag])
                out.update({c[i].tag: list_out})
            else:
                out.update(_node)
        return {in_node.tag: out}

    else:
        if isinstance(in_node.text, str):
            try:
                out = {in_node.tag: float(in_node.text)}
            except ValueError:
                try:
                    out = {in_node.tag: np.array(in_node.text.split(' '), dtype=float)}
                except ValueError:
                    out = {in_node.tag: in_node.text}
            return out
        else:
            return {in_node.tag: ''}



import pandas as pd
from peegy.io.eeg import reader
from os import listdir, walk
from os.path import isfile, isdir, join, basename
import numpy as np
from prettytable import PrettyTable
import logging
import datetime
from peegy.io.external_tools.aep_gui import extsys_tools as et
from peegy.processing.events import event_tools as tt
from peegy.io.external_tools.file_tools import DataLinks
from peegy.io.external_tools.file_tools import match_keywords, flatten_dictionary
from itertools import chain
import os
import glob
import json
__author__ = 'jaime undurraga'


def find_bdf_directories(root_path, f_type='.bdf'):
    out = []
    # search on root path
    root_data_files = [join(root_path, f) for f in listdir(root_path) if
                       isfile(join(root_path, f)) and f.endswith(f_type)]
    if root_data_files:
        out.append(root_path)
    # search in subdirectories
    _directories = [join(root_path, d) for d in listdir(root_path) if isdir(join(root_path, d))]
    for _dir in _directories:
        _out = find_bdf_directories(_dir)
        if _out:
            [out.append(_path) for _path in _out]
        data_files = [join(_dir, f) for f in listdir(_dir) if isfile(join(_dir, f)) and f.endswith(f_type)]
        if data_files:
            out.append(_dir)

    return set(out)


def events_summary(events: np.array([tt.SingleEvent]) = None):
    _codes = np.array([_e.code for _e in events])
    unique_events = np.unique(_codes)
    event_counter = []
    for i, code in enumerate(unique_events):
        event_counter.append({'code': code, 'n': len(np.where(_codes == code)[0])})
    logging.info("\n".join(['Trigger events:', str(event_counter)]))
    print("\n".join(['Trigger events:', str(event_counter)]))


def get_files_and_meta_data(measurement_path: str = '',
                            split_trigger_code: float = None,
                            filter_key_words_in_path: [str] = None,
                            exclude_key_words_in_path: [str] = None,
                            filter_key_words_in_file: [str] = None,
                            exclude_key_words_in_file: [str] = None,
                            deep_match_in_filter_path: bool = True,
                            deep_match_in_filter_file: bool = True,
                            deep_match_in_exclude_path: bool = True,
                            deep_match_in_exclude_file: bool = True,
                            meta_data_file_extension: [str] = ['.json'],
                            meta_data_reference_extension: str = '.extsys'
                            ) -> pd.DataFrame:
    """
    This function search for .bdf or .edf  and .extsys files recursively and try to pair them according to their date.
    The result is a list of DataLinks object, containing the information that links both files.
    :param measurement_path: the root path to search for pairs of files
    :param split_trigger_code: if provided, each .extsys file will be matched to the same bdf file, the ini_time and
    end_time of the DataLinks files will be determined by the trigger_code passed.
    :param filter_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
    include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_path: if provided, the return list will only contain DataLinks object whose paths does
    not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param filter_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
     does include
    the filter_key_words_in_file. This is useful to process entire folders with different conditions
    :param exclude_key_words_in_file: if provided, the return list will only contain DataLinks object whose file name
    does not include
    the exclude_key_words_in_path. This is useful to process entire folders with different conditions
    :param deep_match_in_filter_path: if True, all keywords in deep_match_in_filter_path must be present to return True
    :param deep_match_in_filter_file: if True, all keywords in filter_key_words_in_file must be present to return True
    :param deep_match_in_exclude_path: if True, all keywords in exclude_key_words_in_path must be present to return True
    :param deep_match_in_exclude_file: if True, all keywords in exclude_key_words_in_file must be present to return True
    :param meta_data_file_extension: list indicating the extension of associated meta_data files, e.g. ['.csv', .'json']
    :param meta_data_reference_extension: string indicating the file extension used as a reference to match meta_data
    file
    :return: pandas dataframe with columns extracted from .extsys files. The column DataLinks objects contain the object
     linking .bdf or .edf
    """
    par_out = pd.DataFrame()
    _all_directories = [x[0] for x in walk(measurement_path)]
    for _current_directory in _all_directories:
        if filter_key_words_in_path is not None and not match_keywords(
                filter_key_words_in_path, _current_directory,
                deep_match=deep_match_in_filter_path):
            continue
        if exclude_key_words_in_path is not None and match_keywords(
                exclude_key_words_in_path, _current_directory,
                deep_match=deep_match_in_exclude_path):
            continue

        data_files = [join(_current_directory, f) for f in listdir(_current_directory)
                      if isfile(join(_current_directory, f)) and (f.endswith('.bdf') or f.endswith('.edf'))]
        parameter_files = [join(_current_directory, f) for f in listdir(_current_directory)
                           if isfile(join(_current_directory, f)) and f.endswith('.extsys')]

        if not len(data_files):
            continue
        # get and sort parameters dates
        parameters = []
        for i, _file_name in enumerate(parameter_files):
            parameters.append(et.get_measurement_info_from_zip(_file_name))
        dates = [par['Measurement']['MeasurementModule']['Date'] for x, par in enumerate(parameters)]
        idx_par = [_i[0] for _i in sorted(enumerate([datetime.datetime.strptime(_date, '%m-%d-%y-%H-%M-%S-%f')
                                                     for _date in dates]), key=lambda x: x[1])]
        # get and sort bdf files dates
        _data = []
        for i, _file_name in enumerate(data_files):
            _eeg_reader = reader.eeg_reader(_file_name)
            _data.append(_eeg_reader)
        file_dates = [_c_data._header['start_date'] + '.' + _c_data._header['start_time']
                      for x, _c_data in enumerate(_data)]
        idx_bdf = [_i[0] for _i in sorted(enumerate([datetime.datetime.strptime(_date, '%d.%m.%y.%H.%M.%S')
                                                     for _date in file_dates]), key=lambda x: x[1])]
        # initialize ini and end time to read raw data
        ini_time, end_time = [0] * len(idx_par), [None] * len(idx_bdf)
        if len(idx_bdf) == 1 and split_trigger_code is not None:
            # ensure each parameter file and data file have are paired
            idx_bdf = idx_bdf * len(idx_par)
            raw_events = _data[0].get_events()
            events = tt.detect_events(event_channel=raw_events, fs=_data[0].fs)
            events_summary(events)
            _split_events = np.array([_e for _e in events if _e.code == split_trigger_code])
            assert _split_events.size == len(idx_par), "number of triggers to split conditions does not match" \
                                                       " the number of" \
                                                       " condition files. {:}".format(_data[0].file_name)
            ini_time = [_e.time_pos for _e in _split_events]
            end_time = [_e.time_pos for _e in _split_events[1:]]
            end_time.append(None)

        for i_p, i_b, _ini_time, _end_time in zip(idx_par, idx_bdf, ini_time, end_time):
            _data_file_name, _data_extension = os.path.splitext(_data[i_b].file_name)
            _param_file_name, _param_extension = os.path.splitext(parameter_files[i_p])
            if filter_key_words_in_file is not None and \
                    not (match_keywords(filter_key_words_in_file, _data_file_name,
                                        deep_match=deep_match_in_filter_file) and
                         match_keywords(filter_key_words_in_file, _param_file_name,
                                        deep_match=deep_match_in_filter_file)):
                continue
            if exclude_key_words_in_file is not None and \
                    (match_keywords(exclude_key_words_in_file, _data_file_name,
                                    deep_match=deep_match_in_exclude_file) or
                     match_keywords(exclude_key_words_in_file, _param_file_name,
                                    deep_match=deep_match_in_exclude_file)):
                continue
            if match_keywords(['_bad_'], _param_file_name):
                continue

            # first we check if any file is associated to parameter data file
            if meta_data_reference_extension == _param_extension:
                meta_files = list(chain(*[glob.glob(_param_file_name + '*{:}'.format(_extension))
                                          for _extension in meta_data_file_extension]))
            # if none, we look of any file is associated to data file
            if meta_data_reference_extension == _data_extension:
                meta_files = list(chain(*[glob.glob(_data_file_name + '*{:}'.format(_extension))
                                          for _extension in meta_data_file_extension]))

            meta_data = []
            for _meta_file in meta_files:
                _, file_extension = os.path.splitext(_meta_file)
                if file_extension == '.json':
                    with open(_meta_file, 'r') as f:
                        meta_data.append(json.load(f))

            data_links = DataLinks(parameters=parameters[i_p],
                                   parameters_file=parameter_files[i_p],
                                   parameters_date=dates[i_p],
                                   data_file=_data[i_b].file_name,
                                   data_date=file_dates[i_b],
                                   ini_time=_ini_time,
                                   end_time=_end_time,
                                   meta_data=meta_data
                                   )
            _tmp_pd = pd.DataFrame(parameters[i_p])
            _tmp_pd['Measurement']['data_links'] = data_links
            par_out = pd.concat([par_out, pd.DataFrame.from_dict(
                [flatten_dictionary(dict({k: [v] for k, v in parameters[i_p].items()},
                                         **{'data_links': [data_links]}))])],
                                ignore_index=True)
    if par_out.size:
        _aux = par_out.data_links.values.astype(np.ndarray)
        t = PrettyTable()
        t.add_column(fieldname='File name', column=[basename(x.parameters_file) for x in _aux])
        t.add_column(fieldname='File date', column=[basename(x.parameters_date) for x in _aux])
        t.add_column(fieldname='bdf file name', column=[basename(x.data_file) for x in _aux])
        t.add_column(fieldname='bdf file date', column=[x.data_date for x in _aux])
        logging.info(t)
        print(t)
    return par_out



# -*- coding: utf-8 -*-
"""
Created on Tue Dec  9 10:46:23 2014

@author: jundurraga-ucl
"""
import numpy as np
import matplotlib.pyplot as plt
from peegy.io.eeg.reader import eeg_reader
import os
# import bids

_path = os.path.abspath(os.path.dirname(__file__))
folder_name = os.path.join(_path, "../test_data/set_1/")
# layout = bids.layout.BIDSLayout(folder_name)
# f_name = layout.get(type='eeg', extensions='bdf')[0].filename
f_name = ''
data1 = eeg_reader(file_name=f_name)

triggers = data1.get_triggers()

# retrieve sampling rates (list of sampling rate of each channel)
print("**********************")
print("The sampling rate of, ", f_name, "is", data1.fs, "Hz")
print("--------------\n")

data = data1.get_data()

fig1 = plt.figure(figsize=(7, 6), dpi=100)
ax = fig1.add_subplot(1, 1, 1)
ax.plot(data[:, 0] - np.mean(data[:, 0]))
ax.plot(triggers)
plt.show()



from peegy.processing.tools.epochs_processing_tools import et_average_time_frequency_transformation
import matplotlib.pyplot as plt
import numpy as np
__author__ = 'jaime.undurraga@gmail.com'

# create synthetic data
fs = 44100.0
nsamples = int(np.round(4 * fs))
nchans = 36
ntrials = 100
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs), axis=1)
coeff = np.arange(nchans/2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, -coeff)), 0)

s = source * coeff
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))
# apply jitter
temp_jitter = 1 / f1
for _i in range(s.shape[2]):
    _jitter = int(fs * temp_jitter * (np.random.rand() - 0.5) / 0.5)
    s[:, :, _i] = np.roll(s[:, :, _i], shift=_jitter, axis=0)
SNR = 0.5
noise = np.random.normal(0, 0.1, size=(nsamples, nchans, ntrials))
# noise[1000:12000, :, 30:70] = 20.0 * noise[1000:12000, :, 30:70]
n_source = SNR * s / np.std(s)
data = noise / np.std(noise) + n_source

power_1, time, freqs = et_average_time_frequency_transformation(epochs=data,
                                                                fs=fs,
                                                                time_window=0.5*nsamples/fs,
                                                                sample_interval=0.5*nsamples/fs,
                                                                ave_mode='magnitude'
                                                                )

power_2, time, freqs = et_average_time_frequency_transformation(epochs=data,
                                                                fs=fs,
                                                                time_window=0.5*nsamples/fs,
                                                                sample_interval=0.5*nsamples/fs,
                                                                ave_mode='complex'
                                                                )

thr = 35
_ch = 10
ax = plt.subplot(121)
power_1 = np.abs(power_1)
ref = power_1.max()
power_1 /= ref
p1 = 10*np.log10(power_1)
p1[p1 < -thr] = -thr
cax = ax.pcolormesh(time, freqs, np.squeeze(p1[:, _ch]), cmap=plt.get_cmap('viridis'))
ax.set_xlabel('Time [s]')
ax.set_ylim([0, 60])
ax.set_ylabel('Frequency [Hz]')
ax.set_xlabel('Time [s]')
ax.set_ylim([0, 60])
ax.set_ylabel('Frequency [Hz]')
plt.colorbar(cax)

ax = plt.subplot(122)
power_2 = np.abs(power_2)
power_2 /= ref
p2 = 10*np.log10(power_2)
p2[p2 < -thr] = -thr
cax = ax.pcolormesh(time, freqs, np.squeeze(p2[:, _ch]), cmap=plt.get_cmap('viridis'))
ax.set_xlabel('Time [s]')
ax.set_ylim([0, 60])
ax.set_ylabel('Frequency [Hz]')
plt.colorbar(cax)

plt.show()



"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import epochs_processing_tools as ept
from peegy.processing.tools import eeg_epoch_operators as eop
from peegy.processing.tools.template_generator.auditory_waveforms import abr
from peegy.tools.signal_generator.noise_functions import generate_modulated_noise
import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

# create synthetic data
fs = 8000.0 * u.Hz
duration = 16 * u.ms
n_channels = 4
n_trials = 1200
noise_dim = 4  # dimensionality of noise
source, _ = abr(fs=fs, time_length=duration)
n_samples = source.shape[0]
coeff = np.ones(n_channels//2) * 0.5 / (n_channels / 2)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)
block_size = 100
s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, n_trials))

desired_snr = 5.0
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * n_trials ** 0.5
theoretical_rn = ini_std / n_trials ** 0.5

noise = generate_modulated_noise(fs=fs.value,
                                 duration=source.shape[0] * n_trials / fs.value,
                                 f_noise_low=0,
                                 f_noise_high=4000,
                                 attenuation=0,
                                 n_channels=n_channels) * u.uV

noise = ini_std * noise / np.std(noise, axis=0)
noise = eop.et_fold(noise, n_samples)

# add non-stationary noise
for i in range(3):
    _ini = n_trials//4 + i * 6 * block_size
    noise[:, :, _ini: _ini + block_size] = noise[:, :, _ini: _ini + block_size] * 5

s[:, 0] = s[:, 0] * 0.5
data = noise + s
w_ave, w_w, rn_w, cumulative_rn_w, w_fft, nw, *_ = ept.et_mean(epochs=data,
                                                               block_size=block_size,
                                                               samples_distance=10,
                                                               weighted=True)

w_ave_2, w_w_2, rn_w_2, cumulative_rn_w_2, w_fft_2, nw_2, *_ = ept.et_mean(epochs=data,
                                                                           weight_across_epochs=False,
                                                                           weighted=True)

s_ave, w_s, rn_s, cumulative_rn_s, s_fft, ns, *_ = ept.et_mean(epochs=data,
                                                               block_size=block_size,
                                                               samples_distance=10,
                                                               weighted=False)


snr_w, s_var_w = ept.et_snr_in_rois(data=w_ave, rn=rn_w)
snr_w_2, s_var_w_2 = ept.et_snr_in_rois(data=w_ave_2, rn=rn_w_2)
snr_s, s_var_s = ept.et_snr_in_rois(data=s_ave, rn=rn_s)
print('snr_w in dB = {:}'.format(10 * np.log10(snr_w)))
print('snr_w_2 in dB = {:}'.format(10 * np.log10(snr_w_2)))
print('snr_s in dB = {:}'.format(10 * np.log10(snr_s)))

colour_w = 'red'
colour_w_2 = 'orange'
colour_s = 'blue'

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(w_w[0, :, :].T, color=colour_w)
ax.plot(w_w_2[0, :, :].T, color=colour_w_2)
ax.plot(w_s[0, :, :].T, color=colour_s)
ax.set_ylabel("W ")
ax.set_xlabel("Block number")
plt.show()

fig = plt.figure()
ax = fig.add_subplot(111)
ax.set_title('Noise')
ax.plot(np.std(noise, axis=0).T)
ax.set_ylabel("Noise RMS")
ax.set_xlabel("Trial number")
plt.show()


fig = plt.figure()
ax = fig.add_subplot(141)
ax.plot(np.mean(s, axis=2))
ax.set_title('Target signal')

ax = fig.add_subplot(142)
ax.plot(data[:, 0, :], color=colour_w, alpha=0.005)
ax.plot(w_ave, color='black')
ax.set_title('Weighted average')

ax = fig.add_subplot(143)
ax.plot(data[:, 0, :], color=colour_w_2, alpha=0.005)
ax.plot(w_ave_2, color='black')
ax.set_title('Weighted average 2')

ax = fig.add_subplot(144)
ax.plot(data[:, 0, :], color=colour_s, alpha=0.005)
ax.plot(s_ave, color='black')
ax.set_title('Standard average')
plt.show()

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(cumulative_rn_s, color=colour_s)
ax.set_title('Residual noise')
ax.plot(cumulative_rn_w, color=colour_w)
ax.plot(cumulative_rn_w_2, color=colour_w_2)
plt.show()


fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(10 * np.log10(snr_s), label='final snr_s', color=colour_s)
ax.plot(10 * np.log10(snr_w), label='final snr_w', color=colour_w)
ax.plot(10 * np.log10(snr_w_2), label='final snr_w_2', color=colour_w_2)

ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Channel")
ax.axhline(y=desired_snr)
fig.legend()
plt.show()

cum_snr_w = np.array([])
cum_snr_w_2 = np.array([])
cum_snr_s = np.array([])
cum_snr_w_data = np.array([])
cum_snr_s_data = np.array([])

for i in range(data.shape[2]//block_size):
    end_pos = block_size + i*block_size
    sub_set = data[:, :, 0: end_pos]
    _w_ave, _, _rn_w, *_ = ept.et_mean(epochs=sub_set,
                                       block_size=block_size,
                                       samples_distance=10,
                                       weight_across_epochs=True,
                                       weighted=True)
    _w_ave_2, _, _rn_w_2, *_ = ept.et_mean(epochs=sub_set,
                                           weight_across_epochs=False)

    _s_ave, _, _rn_s,  *_ = ept.et_mean(epochs=sub_set,
                                        block_size=block_size,
                                        samples_distance=10,
                                        weighted=False)

    _rn_w_data = np.std(_w_ave - s[:, :, 0], axis=0)
    _rn_w_data_2 = np.std(_w_ave_2 - s[:, :, 0], axis=0)
    _rn_s_data = np.std(_s_ave - s[:, :, 0], axis=0)
    _snr_w, s_var_w = ept.et_snr_in_rois(data=_w_ave, rn=_rn_w)
    _snr_w_2, s_var_w_2 = ept.et_snr_in_rois(data=_w_ave_2, rn=_rn_w_2)
    _snr_s, s_var_s = ept.et_snr_in_rois(data=_s_ave, rn=_rn_s)
    _snr_w_data, s_var_w_data = ept.et_snr_in_rois(data=_w_ave, rn=_rn_w_data)
    _snr_w_2_data, s_var_w_data_2 = ept.et_snr_in_rois(data=_w_ave_2, rn=_rn_w_data_2)
    _snr_s_data, s_var_s_data = ept.et_snr_in_rois(data=_s_ave, rn=_rn_s_data)

    if i == 0:
        cum_snr_w = _snr_w
        cum_snr_w_2 = _snr_w_2
        cum_snr_s = _snr_s
        cum_snr_w_data = _snr_w_data
        cum_snr_w_2_data = _snr_w_2_data
        cum_snr_s_data = _snr_s_data
    else:
        cum_snr_w = np.hstack([cum_snr_w, _snr_w])
        cum_snr_w_2 = np.hstack([cum_snr_w_2, _snr_w_2])
        cum_snr_s = np.hstack([cum_snr_s, _snr_s])
        cum_snr_w_data = np.hstack([cum_snr_w_data, _snr_w_data])
        cum_snr_w_2_data = np.hstack([cum_snr_w_2_data, _snr_w_2_data])
        cum_snr_s_data = np.hstack([cum_snr_s_data, _snr_s_data])

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(10 * np.log10(cum_snr_w.T + 1e-3), color=colour_w)
ax.plot(10 * np.log10(cum_snr_w_2.T + 1e-3), color=colour_w_2)
ax.plot(10 * np.log10(cum_snr_s.T + 1e-3), color=colour_s)
ax.axhline(y=desired_snr)
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")
plt.show()

fig = plt.figure()
ax = fig.add_subplot(321)
ax.plot(10 * np.log10(cum_snr_w.T + 1e-3), color=colour_w)
ax.axhline(y=desired_snr)
ax.set_title("SNRw [dB]")
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")

ax = fig.add_subplot(322)
ax.plot(10 * np.log10(cum_snr_w_data.T + 1e-3), color=colour_w)
ax.axhline(y=desired_snr)
ax.set_title("SNRw real [dB]")
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")

ax = fig.add_subplot(323)
ax.plot(10 * np.log10(cum_snr_w_2.T + 1e-3), color=colour_w_2)
ax.axhline(y=desired_snr)
ax.set_title("SNRw_2 [dB]")
ax.set_ylabel("SNR_2 [dB]")
ax.set_xlabel("Block number")

ax = fig.add_subplot(324)
ax.plot(10 * np.log10(cum_snr_w_2_data.T + 1e-3), color=colour_w_2)
ax.axhline(y=desired_snr)
ax.set_title("SNRw_2 real [dB]")
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")

ax = fig.add_subplot(325)
ax.plot(10 * np.log10(cum_snr_s.T + + 1e-3), color=colour_s)
ax.axhline(y=desired_snr)
ax.set_title("SNRs [dB]")
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")

ax = fig.add_subplot(326)
ax.set_title("SNRs real [dB]")
ax.plot(10 * np.log10(cum_snr_s_data.T + 1e-3), color=colour_s)
ax.axhline(y=desired_snr)
ax.set_ylabel("SNR [dB]")
ax.set_xlabel("Block number")
plt.show()



"""
.. _tut-acc-sim:

Test spatial filtering with evoked response (Simulation)
==============================

This example simulates an evoked responses with non-stationary noise.
We illustrate how weighted averaging and DSS can lead to much better results in terms of recovering the data.
This example follows that by Alain de Cheveigné (http://audition.ens.fr/adc/NoiseTools/) but it has been extended
to include non-stationary noise and weighted average.
.. contents:: Page contents
   :local:
   :depth: 2
"""
import matplotlib.pyplot as plt
from peegy.processing.tools.eeg_epoch_operators import et_mmat
from peegy.processing.tools.epochs_processing_tools import et_mean, et_frequency_mean2
import numpy as np
from scipy.stats import norm
from astropy import units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')
__author__ = 'jundurraga'

# create synthetic data
nsamples = 500 * 3
nchans = 10
ntrials = 300
noise_dim = 20  # dimensionality of noise
freq = 100
fs = 1000

source = np.concatenate((
    np.zeros((nsamples // 3, )),
    np.sin(2 * np.pi * np.arange(nsamples / 3) / fs * freq),
    np.zeros((nsamples // 3, ))))

# source = np.sin(2 * np.pi * np.arange(nsamples) / fs * freq)

np.random.seed(2)
mix_m = norm.ppf(np.random.rand(nchans, 1))
s = source * mix_m
s = np.tile(np.expand_dims(s.T, axis=2), (1, 1, ntrials))
SNR = 0.5

noise_unmixed = norm.ppf(np.random.rand(noise_dim * ntrials * nsamples))
noise_unmixed = np.reshape(noise_unmixed, [nsamples, noise_dim, ntrials], order='F')

noise_mix = norm.ppf(np.random.rand(nchans * noise_dim))
noise_mix = np.reshape(noise_mix, [noise_dim, nchans], order='F')
noise = et_mmat(noise_unmixed, noise_mix)

scaled_source = SNR * s * u.uV / np.std(s)
scaled_noise = noise * u.uV / np.std(noise)
block_size = 20
# generate non-stationary noise
for i in np.arange(1, 4):
    _ini = i*20
    _end = _ini + 10
    scaled_noise[:, :, _ini: _end] += 0.5e2 * i * scaled_noise[:, :, _ini: _end]

# mix data
data = scaled_noise + scaled_source
frequencies = freq * (1 + np.arange(1))

rn_w_vector = []
rn_s_vector = []
real_rn_w_vector = []
real_rn_s_vector = []
snr_w_vector = []
snr_s_vector = []
snr_w_dss_vector = []
real_snr_w_vector = []
real_snr_s_vector = []
ave_source, source_w, source_rn, _, source_fft, n, *_ = et_mean(scaled_source, block_size=block_size)
freq_axis = np.arange(0, ave_source.shape[0]) * fs / ave_source.shape[0]
scaling_factor = 2 / data.shape[0]
_idx_freq = np.argmin(np.abs(freq_axis - freq))
target_amp = np.abs(np.fft.fft(ave_source, axis=0))[_idx_freq, :] * scaling_factor
delta_freq = 3
for _n in np.arange(block_size, data.shape[2], block_size):
    subset = np.copy(data)[:, :, 0: _n]

    # standard averaging of the data
    s_ave, s_snr, s_rn, s_rn_fre, s_snr_freq, s_fft, w_s, _, _ = et_frequency_mean2(subset,
                                                                                    fs=fs,
                                                                                    block_size=block_size,
                                                                                    weighted_average=False,
                                                                                    test_frequencies=frequencies,
                                                                                    delta_frequency=delta_freq)
    # weighted averaging of the data
    w_ave, w_snr, w_rn, w_rn_freq, w_snr_freq, w_fft, w_w, _, _ = et_frequency_mean2(subset,
                                                                                     fs=fs,
                                                                                     block_size=block_size,
                                                                                     weighted_average=True,
                                                                                     test_frequencies=frequencies,
                                                                                     delta_frequency=delta_freq)

    # compute actual residual noise by subtracting the source signal
    real_noise_standard_ave = ave_source - s_ave
    real_noise_weighted_ave = ave_source - w_ave

    fft_noise_standard_ave = np.abs(np.fft.fft(real_noise_standard_ave, axis=0)) * scaling_factor
    fft_noise_weighted_ave = np.abs(np.fft.fft(real_noise_weighted_ave, axis=0)) * scaling_factor

    rn_s_vector.append(s_rn)
    rn_w_vector.append(w_rn)

    _real_noise_s_ave = fft_noise_standard_ave[_idx_freq, :]
    _real_noise_w_ave = fft_noise_weighted_ave[_idx_freq, :]

    real_rn_s_vector.append(_real_noise_s_ave)
    real_rn_w_vector.append(_real_noise_w_ave)

    snr_s_vector.append(s_snr)
    snr_w_vector.append(w_snr)

    real_snr_s_vector.append(target_amp ** 2 / _real_noise_s_ave ** 2)
    real_snr_w_vector.append(target_amp ** 2 / _real_noise_w_ave ** 2)

# plot noise
plt.figure()
plt.plot(real_rn_s_vector, 'k')
plt.plot(rn_s_vector, 'b')

plt.figure()
plt.plot(real_rn_w_vector, 'k')
plt.plot(rn_w_vector, 'b')


# plot snr
n_epochs = np.arange(block_size, data.shape[2], block_size)
plt.figure()
plt.plot(n_epochs, 10 * np.log10(real_snr_s_vector), 'k')
plt.plot(n_epochs, 10 * np.log10(snr_s_vector), 'b')

noise_sample = scaled_noise[0, 0, :]
plt.plot(noise_sample / np.abs(noise_sample).max())
w_sample = w_s[0, 0, :]
plt.plot(w_sample / np.max(w_sample), 'g')

plt.figure()
plt.plot(n_epochs, 10 * np.log10(real_snr_w_vector), 'k')
plt.plot(n_epochs, 10 * np.log10(snr_w_vector), 'b')
plt.plot(noise_sample / np.abs(noise_sample).max())
w_sample = w_w[0, 0, :]
plt.plot(w_sample / np.max(w_sample), 'g')

plt.show()



import copy
import pandas
import tqdm
# from joblib import Parallel, delayed
from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.pipe.general import RemoveEvokedResponse
from peegy.processing.pipe.pipeline import PipePool
from peegy.definitions.tables import Tables
from peegy.processing.statistics.definitions import TestType
import numpy as np
import pandas as pd
import astropy.units as u


class BootstrapTarget(object):
    """
    This class is used to define the target tables that will be taking for bootstrapping.
    """
    def __init__(self,
                 table_name: str = None,
                 group_by: [str] = None,
                 target_values: [str] = None
                 ):
        """

        :param table_name: What table should be targeted for bootstrapping
        :param group_by: column names (present in the target statistical table) that should be used to group
        the data that should be bootstrapped.
        :param target_values: column names with the statistics that should be bootstrapped (e.g. f)
        """
        self.table_name = table_name
        self.group_by = group_by
        self.table_name = table_name
        self.target_values = target_values


class Bootstrap(InputOutputProcess):
    def __init__(self,
                 origin_input_process: InputOutputProcess = None,
                 before_bootstrapping_input_process: InputOutputProcess = None,
                 at_each_bootstrap_do: PipePool = None,
                 n_events: int = None,
                 event_code: float = 1,
                 dur: u.Quantity = 100 * u.us,
                 min_time: u.Quantity = None,
                 max_time: u.Quantity = None,
                 epoch_length: u.Quantity = None,
                 n_bootstraps: int = 1000,
                 bootstrap_targets: [BootstrapTarget] = None,
                 alpha_level: float = 0.05,
                 one_sided: bool = True,
                 **kwargs) -> InputOutputProcess:
        """
        Bootstrap input pipeline. It assumed that the origin_input_process contains the continuous, un-epoched data.
        The events will be randomized and the pipeline passed in at_each_bootstrap_do will be run.
        The process of shuffling events and running the pipeline  will be repeated n_bootstraps times.
        The output table will contain the mean, median, standard deviation, lower confidence interval, upper confidence
        interval, sample size for the bootstrap and the p-value.
        Confidence intervals are based on the alpha_value passed.
        :param origin_input_process: The continuous, un-epoched data that will be bootstrap
        :param before_bootstrapping_input_process: Inputprocess to be run before bootstrap begins but after the
        original statistics have been computed. If None, by default the average evoked response will be removed.
        :param at_each_bootstrap_do: The pipeline that will be bootstrapped
        :param n_events: If not none, the number of random events will be n_events
        :param event_code: the event code that will be shuffled
        :param dur: duration of the shuffled event code
        :param min_time: minimum time allowed for shuffle events. If None, the timing of the first event will be used
        :param max_time: maximum time allowed for shuffle events. If None, the timing of the last event will be used
        :param epoch_length: reference duration of epochs. This is only used to ensure that the after the last event
        there will be time enough to ensure this epoch length
        :param n_bootstraps: how many times to bootstrap the pipeline
        :param bootstrap_targets: array of BootstrapTarget indicating what tables from the different pipeline elements
        should be keep for bootstrapping and what statistics.
        :param alpha_level: alpha level of confidence intervals
        :param one_sided: bool indicating whether alpha value is one_sided or not.
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(Bootstrap, self).__init__(input_process=None, **kwargs)
        self.origin_input_process = origin_input_process
        self.before_bootstrapping_input_process = before_bootstrapping_input_process
        self.at_each_bootstrap_do = at_each_bootstrap_do
        self.n_bootstraps = n_bootstraps
        self.n_events = n_events
        self.event_code = event_code
        self.dur = dur
        self.min_time = min_time
        self.max_time = max_time
        self.epoch_length = epoch_length
        self.alpha_level = alpha_level
        self.one_sided = one_sided
        self._original_events = None
        self.bootstrap_targets = bootstrap_targets

    def transform_data(self):
        # the origin input_process is copied to ensure nothing is done to the original reference process
        if self.origin_input_process is not None:
            _origin_input_process = copy.copy(self.origin_input_process)
        else:
            _origin_input_process = copy.copy(list(self.at_each_bootstrap_do.values())[0].input_process)

        # ensure that the reference process is used by the pipeline
        for _process in list(self.at_each_bootstrap_do.values()):
            if _process.input_process.name == _origin_input_process.name:
                _process = _origin_input_process

        # first run pipeline to extract actual values to be tested against bootstrapped tables
        self.at_each_bootstrap_do.run(force=True)
        # extract reference tables
        ref_table = get_tables(pipeline=self.at_each_bootstrap_do, targets=self.bootstrap_targets)
        # get event information
        _n_events = self.n_events
        _min_time = self.min_time
        _max_time = self.max_time
        if _n_events is None:
            _n_events = _origin_input_process.output_node.events.get_events_code(code=self.event_code).size
        if _min_time is None:
            _min_time = _origin_input_process.output_node.events.get_events_time(code=self.event_code).min()
        if _max_time is None:
            _max_time = _origin_input_process.output_node.events.get_events_time(code=self.event_code).max()
        # apply whatever we ask before we bootstrap. Normally we will remove the evoked data
        if self.before_bootstrapping_input_process is not None:
            self.before_bootstrapping_input_process.input_process = _origin_input_process
        else:
            self.before_bootstrapping_input_process = RemoveEvokedResponse(
                input_process=_origin_input_process,
                event_code=self.event_code,
                post_stimulus_interval=self.epoch_length)

        self.before_bootstrapping_input_process.run()
        _origin_input_process.output_node.data = self.before_bootstrapping_input_process.output_node.data
        # generate average epoch length if not given to prevent small estimations from random events
        _epoch_length = self.epoch_length
        if self.epoch_length is None:
            _epoch_length = np.mean(
                np.diff(_origin_input_process.output_node.events.get_events_time(code=self.event_code)))

        # bootstrap pipeline
        bootstrapped_tables = pd.DataFrame()
        for _i in tqdm.tqdm(range(self.n_bootstraps), desc='Bootstrapping pipeline', colour='blue'):
            # randomize events for run
            _origin_input_process.output_node.randomize_events(event_code=self.event_code,
                                                               epoch_length=_epoch_length,
                                                               min_time=_min_time,
                                                               max_time=_max_time,
                                                               n_events=_n_events
                                                               )
            # run pipeline with new events
            self.at_each_bootstrap_do.run(force=True)
            if _i == 0:
                bootstrapped_tables = get_tables(pipeline=self.at_each_bootstrap_do, targets=self.bootstrap_targets)
                bootstrapped_tables['bootstrap_number'] = _i

            else:
                _current_tables = get_tables(pipeline=self.at_each_bootstrap_do, targets=self.bootstrap_targets)
                _current_tables['bootstrap_number'] = _i
                bootstrapped_tables = pd.concat([bootstrapped_tables, _current_tables])
        # compute statistics
        ref_table = ref_table.rename(columns={'value': 'statistic_value'})
        columns_left = set(bootstrapped_tables.keys()).difference(['value', 'bootstrap_number'])
        columns_right = set(ref_table.keys()).difference(['statistic_value'])
        common_cols = list(columns_left.intersection(columns_right))
        bootstrapped_and_ref_tables = pd.merge(bootstrapped_tables,
                                               ref_table,
                                               how='left',
                                               on=common_cols
                                               )

        df_boots = bootstrapped_and_ref_tables.groupby(list(set(bootstrapped_and_ref_tables.keys()).difference(
            ['value', 'bootstrap_number'])), as_index=False).apply(
            lambda x: pd.Series({'stats': get_bootstrap_estimates(samples=x['value'],
                                                                  statistic=x['statistic_value'],
                                                                  alpha_level=self.alpha_level,
                                                                  one_sided=self.one_sided)}))

        (df_boots['mean'],
         df_boots['median'],
         df_boots['std'],
         df_boots['ci_lower'],
         df_boots['ci_upper'],
         df_boots['sample_size'],
         df_boots['p_value']) = zip(*df_boots.stats)
        df_boots['one_sided'] = self.one_sided
        df_boots = df_boots.drop(['stats'], axis=1)
        # join boot table and reference table
        self.output_node.statistical_tests = Tables(table_name=TestType.bootstrap,
                                                    data=df_boots,
                                                    data_source=self.name)


def get_bootstrap_estimates(samples: np.array = None,
                            statistic: np.array = None,
                            alpha_level: float = 0.05,
                            one_sided=True):
    # remove nan that may be present
    samples = samples[~pd.isna(samples)]
    _stat = np.unique(statistic)
    sample_size = samples.shape[0]
    sorted_samples = np.sort(samples, axis=0)
    mean = np.mean(samples, axis=0)
    median = np.median(samples, axis=0)
    std = np.std(samples, axis=0)
    _alpha = alpha_level
    if not one_sided:
        _alpha = alpha_level / 2.0

    _i_upper = np.minimum(int(samples.shape[0] * (1 - _alpha)), samples.shape[0])
    _i_lower = np.maximum(int(samples.shape[0] * _alpha), 0)
    ci_lower = sorted_samples[_i_lower]
    ci_upper = sorted_samples[_i_upper]
    diff = np.abs(_stat - sorted_samples)
    _idx_match = np.argmin(diff, axis=0) + 1
    p_value = 1 - _idx_match / sample_size
    return [mean, median, std, ci_lower, ci_upper, sample_size, p_value]


def get_tables(pipeline: PipePool = None,
               targets: [BootstrapTarget] = None):
    output = pd.DataFrame()
    for _source, _value in pipeline.items():
        for _target in targets:
            if _target.table_name in _value.output_node.statistical_tests.keys():
                _columns = list(set(_value.output_node.statistical_tests[_target.table_name].keys()) &
                                set(_target.target_values))
                if len(_columns):
                    df = _value.output_node.statistical_tests[_target.table_name]
                    df = df[_target.group_by + _target.target_values + ['data_source']]
                    df = pd.melt(df, ignore_index=False,
                                 id_vars=set(set(df.keys())).difference(_target.target_values),
                                 value_vars=_target.target_values,
                                 var_name='statistic')
                    df['table_name'] = _target.table_name
                    output = pandas.concat([output, df], ignore_index=True)
    return output



import matplotlib.pyplot as plt
from peegy.processing.tools.epochs_processing_tools import bootstrap
import numpy as np
__author__ = 'jundurraga-ucl'

# create synthetic data
fs = 44100.0
nsamples = int(np.round(1 * fs))
nchans = 16
ntrials = 100
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs), axis=1)
coeff = np.arange(nchans/2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, -coeff)), 0)

s = source * coeff
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))
SNR = 6
noise = np.random.normal(0, 0.1, size=(nsamples, nchans, ntrials))
n_source = SNR * s / np.std(s)
data = noise / np.std(noise) + n_source

mean, ci_low, ci_high = bootstrap(data=data, num_samples=1000)

ax = plt.subplot(111)
ax.plot(np.mean(n_source, axis=2)[:, 10])
ax = plt.subplot(111)
ax.plot(np.mean(data, axis=2)[:, 10])
ax = plt.subplot(111)
ax.plot(mean[:, 10])
ax.fill_between(x=np.arange(0, mean.shape[0]), y1=mean[:, 10], y2=ci_high[:, 10])
ax.fill_between(x=np.arange(0, mean.shape[0]), y1=ci_low[:, 10], y2=mean[:, 10])

plt.show()



import copy
from tqdm import tqdm
from peegy.processing.pipe.definitions import InputOutputProcess
import matplotlib.pyplot as plt
from peegy.plot import eeg_ave_epochs_plot_tools as eegpt
from peegy.processing.tools.video.recording import Recorder
import numpy as np
import os
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
import gc
from PyQt5.QtCore import QLibraryInfo
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class PlotWaveforms(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 overlay: [InputOutputProcess] = None,
                 ch_to_plot: np.array = None,
                 title: str = '',
                 plot_x_lim: [float, float] = None,
                 plot_y_lim: [float, float] = None,
                 offset_step: u.quantity.Quantity = None,
                 return_figures: bool = False,
                 statistical_test: str = None,
                 show_following_stats: [str] = None,
                 fig_format: str = '.png',
                 fontsize: float = 12,
                 user_naming_rule: str = '',
                 save_to_file: bool = True,
                 show_peaks: bool = True,
                 show_labels: bool = True,
                 **kwargs):
        """
        This InputOutputProcess plots all channel waveforms in a single plot and saves them.
        :param input_process: an InputOutput process
        :param ch_to_plot: numpy array with index or labels of channels to plot
        :param overlay: list of InputOutputProcess to overlay
        :param title: title of the plot
        :param plot_x_lim: x limits of the plot
        :param plot_y_lim: y limits of the plot
        :param offset_step: offset between different channels
        :param return_figures: If true, handle to figure will be passed to self.figures
        :param statistical_test: name of statistical test to extract information.
        :param show_following_stats: list of string indicating which parameters found in statistical_tests are shown
        :param fig_format: string indicating the format of the output figure (e.g. '.png' or '.pdf')
        :param fontsize: size of fonts in plot
        :param idx_channels_to_plot: np.array indicating the index of the channels to plot
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param save_to_file: if True, figure will be saved
        :param show_peaks: if True, detected peaks will be shown
        :param show_labels: if True, peak labels will be shown
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(PlotWaveforms, self).__init__(input_process=input_process, **kwargs)
        self.ch_to_plot = ch_to_plot
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.offset_step = set_default_unit(offset_step, u.uV)
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.overlay = overlay
        self.statistical_test = statistical_test
        self.show_following_stats = show_following_stats
        self.return_figures = return_figures
        self.save_to_file = save_to_file
        self.show_peaks = show_peaks
        self.show_labels = show_labels

    def transform_data(self):
        figure_dir_path = self.input_node.paths.figures_current_dir
        _sep = '_' if self.user_naming_rule is not None else ''
        figure_basename = self.input_process.name + _sep + self.user_naming_rule
        to_plot = [self.input_node]
        if self.overlay is not None:
            to_plot += [_in.output_node for _in in self.overlay]

        figure = eegpt.plot_single_channels(ave_data=to_plot,
                                            channels=self.ch_to_plot,
                                            figure_dir_path=figure_dir_path,
                                            figure_basename=figure_basename,
                                            title=self.title,
                                            x_lim=self.plot_x_lim,
                                            y_lim=self.plot_y_lim,
                                            offset_step=self.offset_step,
                                            fig_format=self.fig_format,
                                            fontsize=self.fontsize,
                                            statistical_test=self.statistical_test,
                                            show_following_stats=self.show_following_stats,
                                            save_to_file=self.save_to_file,
                                            show_peaks=self.show_peaks,
                                            show_labels=self.show_labels
                                            )
        if self.return_figures:
            self.figures = figure
        else:
            plt.close(figure)
            gc.collect()


class PlotTopographicMap(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels: np.array([str]) = np.array([]),
                 title: str = '',
                 plot_x_lim: list = None,
                 plot_y_lim: list = None,
                 times: np.array = np.array([]),
                 save_to_file: bool = True,
                 fig_format: str = '.png',
                 fontsize: float = 12,
                 user_naming_rule: str = '',
                 return_figures: bool = False,
                 **kwargs
                 ):
        """
        Plot topographic maps of figures at desired times. If the input_process contains time or frequency peaks from
        PeakDetectionTimeDomain or tests such as FTest, AverageEpochsFrequencyDomain, or PhaseLockingValue,
        topographic maps for those peaks will be also shown
        :param input_process: an InputOutput process
        :param topographic_channels: a numpy array (integers index or string with channel labels) indicating channels to
        be plotted
        :param title: string with the desired figure title
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param times: numeric array with times to show topographic maps
        :param save_to_file: if True, figure will be saved
        :param fig_format: string indicating the format of the figure being saved
        :param fontsize: float indicating the size of the fonts
        :param user_naming_rule: string containing an extra text to be included in the name of the figure
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(PlotTopographicMap, self).__init__(input_process=input_process, **kwargs)
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = set_default_unit(times, u.s)
        self.return_figures = return_figures
        self.save_to_file = save_to_file

    def transform_data(self):
        _sep = '_' if self.user_naming_rule is not None else ''
        _figure_base_name = self.input_process.name + _sep + self.user_naming_rule
        figures = eegpt.plot_eeg_topographic_map(ave_data=self.input_node,
                                                 eeg_topographic_map_channels=self.topographic_channels,
                                                 figure_dir_path=self.input_node.paths.figures_current_dir,
                                                 figure_basename=_figure_base_name,
                                                 times=self.times,
                                                 domain=self.input_node.domain,
                                                 title=self.title,
                                                 x_lim=self.plot_x_lim,
                                                 y_lim=self.plot_y_lim,
                                                 fig_format=self.fig_format,
                                                 fontsize=self.fontsize,
                                                 return_figures=self.return_figures,
                                                 save_figures=self.save_to_file)
        if self.return_figures:
            self.figures = figures
        else:
            [plt.close(_fig) for _fig in figures]
            gc.collect()


class PlotSpectrogram(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels=np.array([]),
                 title='',
                 plot_x_lim=None,
                 plot_y_lim=None,
                 times=np.array([]),
                 fig_format='.png',
                 fontsize=12,
                 user_naming_rule='',
                 time_window=2.0,
                 sample_interval=0.004,
                 spec_thresh=4
                 ):
        super(PlotSpectrogram, self).__init__(input_process=input_process)
        self.topographic_channels = topographic_channels
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule
        self.times = times
        self.time_window = time_window
        self.sample_interval = sample_interval
        self.spec_thresh = spec_thresh

    def transform_data(self):
        eegpt.plot_eeg_time_frequency_transformation(ave_data=self.input_node,
                                                     eeg_topographic_map_channels=self.topographic_channels,
                                                     figure_dir_path=self.input_node.paths.figure_basename_path,
                                                     figure_basename='{:}{:}'.format(
                                                         self.input_process.name + '_',
                                                         self.user_naming_rule),
                                                     time_unit='s',
                                                     amplitude_unit='uV',
                                                     times=self.times,
                                                     domain=self.input_node.domain,
                                                     title=self.title,
                                                     x_lim=self.plot_x_lim,
                                                     y_lim=self.plot_y_lim,
                                                     fig_format=self.fig_format,
                                                     fontsize=self.fontsize,
                                                     time_window=self.time_window,
                                                     sample_interval=self.sample_interval,
                                                     spec_thresh=self.spec_thresh
                                                     )


class TopographicMapVideo(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 topographic_channels: np.array([str]) = np.array([]),
                 step_size: u.Quantity = None,
                 title: str = '',
                 plot_x_lim: list = None,
                 plot_y_lim: list = None,
                 fig_format: str = '.png',
                 fontsize: float = 12,
                 user_naming_rule: str = '',
                 **kwargs
                 ):
        """
        Generate a video with the topographic over time. If the input_process contains time or frequency peaks from
        PeakDetectionTimeDomain or tests such as FTest, AverageEpochsFrequencyDomain, or PhaseLockingValue,
        topographic maps for those peaks will be also shown
        :param input_process: an InputOutput process
        :param topographic_channels: a numpy array (integers index or string with channel labels) indicating channels to
        be plotted
        :param step_size: step size to generate video in same units as self.input_node.x
        :param title: string with the desired figure title
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param times: numeric array with times to show topographic maps
        :param fig_format: string indicating the format of the figure being saved
        :param fontsize: float indicating the size of the fonts
        :param user_naming_rule: string containing an extra text to be included in the name of the figure
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(TopographicMapVideo, self).__init__(input_process=input_process, **kwargs)
        self.topographic_channels = topographic_channels
        self.step_size = step_size
        self.title = title
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.fontsize = fontsize
        self.user_naming_rule = user_naming_rule

    def transform_data(self):
        self.step_size = set_default_unit(self.step_size, self.input_node.x.unit)
        self.plot_x_lim = set_default_unit(self.plot_x_lim, self.input_node.x.unit)
        self.plot_y_lim = set_default_unit(self.plot_y_lim, self.input_node.data.unit)

        data_node = copy.copy(self.input_node)
        data_node.peak_times = None
        data_node.peak_frequency = None
        # Generate a video showing the topographic map over time
        _x_values = self.input_node.x
        if self.step_size is not None:
            skip = round((self.step_size * self.input_node.fs).to_value(u.dimensionless_unscaled))
            _x_values = _x_values[0::skip]
        for _ch in self.topographic_channels:
            _video_path = self.input_node.paths.figures_current_dir
            _sep = '_' if self.user_naming_rule is not None else ''
            _video_base_name = self.input_process.name + _sep + self.user_naming_rule
            rec = Recorder(video_path=_video_path + _video_base_name + '_' + _ch + '.mp4')
            for _x in tqdm(_x_values, desc='Generating Video for channel {:}'.format(_ch)):
                plt.ioff()
                figures = eegpt.plot_eeg_topographic_map(ave_data=data_node,
                                                         eeg_topographic_map_channels=[_ch],
                                                         save_figures=False,
                                                         times=_x,
                                                         domain=self.input_node.domain,
                                                         title=self.title,
                                                         x_lim=self.plot_x_lim,
                                                         y_lim=self.plot_y_lim,
                                                         fig_format=self.fig_format,
                                                         fontsize=self.fontsize,
                                                         return_figures=True)
                if figures is not None:
                    rec.add_frame(figures[0])
                    for _fig in figures:
                        plt.close(_fig)
                    gc.collect()
                # print_progress_bar(iteration=_i, total=x.size, prefix='Generating Video for channel {:}'.format(_ch))
            rec.end_recording()



from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.detection.time_domain_tools import detect_peaks_and_amplitudes, TimePeakWindow, \
    PeakToPeakMeasure
import numpy as np
import pandas as pd
import os
from peegy.definitions.tables import Tables
from PyQt5.QtCore import QLibraryInfo
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class PeakDetectionTimeDomain(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess, time_peak_windows=np.array([TimePeakWindow]),
                 peak_to_peak_measures=np.array([PeakToPeakMeasure]), **kwargs):
        super(PeakDetectionTimeDomain, self).__init__(input_process=input_process, **kwargs)
        self.time_peak_windows = time_peak_windows
        self.peak_to_peak_measures = peak_to_peak_measures

    def transform_data(self):
        peak_containers, amplitudes = detect_peaks_and_amplitudes(
            data_node=self.input_node,
            time_peak_windows=self.time_peak_windows,
            eeg_peak_to_peak_measures=self.peak_to_peak_measures)

        _data_pd = pd.concat([_peaks.to_pandas() for _peaks in peak_containers], ignore_index=True)
        _amps_pd = pd.concat([_amp.to_pandas() for _amp in amplitudes], ignore_index=True)
        self.output_node.data = self.input_node.data
        self.output_node.peaks = _data_pd
        self.output_node.processing_tables_local = Tables(table_name='peaks_time',
                                                          data=_data_pd,
                                                          data_source=self.name)
        self.output_node.processing_tables_local.append(Tables(table_name='amplitudes',
                                                               data=_amps_pd,
                                                               data_source=self.name))
        self.output_node.statistical_tests = self.input_process.output_node.statistical_tests



import numpy as np
from peegy.processing.tools.filters.resampling import eeg_resampling
import matplotlib.pyplot as plt
import astropy.units as u
__author__ = 'jundurraga-ucl'


fs = 1000.0 * u.Hz
f_s = 50.0 * u.Hz
n_o = 700.0
t = np.arange(0, n_o) / fs
y1 = np.sin(2 * np.pi * u.rad * f_s * t + np.pi * u.rad / 2) * u.uV
y2 = np.sin(2 * np.pi * u.rad * 4 * f_s * t) * u.uV
y = np.tile(np.array(y1 + y2) + 0.0 * np.random.random(t.shape), (4, 1)).T * u.uV
new_fs = 2500 * u.Hz

yf = np.fft.fft(y, axis=0)
freq = np.arange(len(yf)) * fs / len(yf)
yc, factor = eeg_resampling(x=y, new_fs=new_fs, fs=fs, blocks=8)
new_fs = fs * factor
new_time = np.arange(0, yc.shape[0]) / new_fs
plt.plot(t, y, label='original', color='b')
plt.plot(t, y, 'bo')
plt.plot(t, y1, label='h1')
plt.plot(t, y2, label='h1')
plt.plot(np.squeeze(new_time), np.squeeze(yc), 'r', label='resampled')
plt.plot(np.squeeze(new_time), np.squeeze(yc), 'rv', label='resampled')
plt.legend()
plt.show()
plt.show()



import copy
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import matplotlib.gridspec as gridspec
from peegy.layouts.layouts import Layout
from peegy.plot import eeg_plot_tools as eegpt
from matplotlib import ticker
from astropy import units as u

__author__ = 'jundurraga'


def plot_time_frequency_responses(dataframe: pd.DataFrame = None,
                                  rows_by: str = None,
                                  cols_by: str = None,
                                  sub_average_time_buffer_size: int = None,
                                  blend_subcategories: bool = False,
                                  time_xlim: [float, float] = None,
                                  time_ylim: [float, float] = None,
                                  freq_xlim: [float, float] = None,
                                  freq_ylim: [float, float] = None,
                                  time_vmarkers: np.array = None,
                                  freq_vmarkers: np.array = None,
                                  freq_vmarkers_style: str = None,
                                  show_individual_waveforms: bool = True,
                                  individual_waveforms_alpha: float = 0.1,
                                  show_sem: bool = True,
                                  sem_alpha: float = 0.2,
                                  show_sd: bool = True,
                                  sd_alpha: float = 0.2,
                                  show_legend: bool = True,
                                  title_by: str = 'row',
                                  title_v_offset: float = 0.0,
                                  ylabel: str = None,
                                  y_unit_to: u.Unit = None,
                                  x_unit_to: u.Unit = None,
                                  fontsize: float = 6,
                                  ) -> plt.figure:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data).
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy
    :param rows_by: name of the factor in the dataframe for which the rows in the plot will be split
    :param cols_by: name of the factor in the dataframe for which the columns in the plot will be split
    :param sub_average_time_buffer_size: This is a parameter used to sub_average time_domain data. For example, if each
    of your data have 10000 points, and you want to show the average having a length of 1000 samples, you could specify
    sub_average_time_buffer_size = 1000. This will averaged the 10000 points by splitting the data into blocks of 1000
    samples
    :param blend_subcategories: if true, any subcategory present in each column and row will be pull together as
    belonging to the same category
    :param time_xlim: x axis limis for the time-domain panels
    :param time_ylim: y axis limis for the time-domain panels
    :param freq_xlim: x axis limis for the frequency-domain panels
    :param freq_ylim: y axis limis for the frequency-domain panels
    :param time_vmarkers: array with x values to add a vertical marker in the time-domain panels
    :param freq_vmarkers: array with x values to add a vertical marker in the frequency-domain panels
    :param show_individual_waveforms: if true, individual waveforms will be shown.
    :param individual_waveforms_alpha: value between 0 and 1 indicating the alpha level of individual waveforms
    :param show_sem: if true, the standard error of the mean will be shown
    :param sem_alpha: value between 0 and 1 indicating the alpha level of the sem
    :param show_sd: if true, the standard deviation of the mean will be shown
    :param sd_alpha: value between 0 and 1 indicating the alpha level of the standard deviation
    :param show_legend: if True, the legend for any other category present in the dataframe will be shown
    :param title_by: string specifying from which factor you want to show the titles in each panel. This can be: "row",
    "col", or "both"
    :param title_v_offset: float specifying the vertical offset of the title
    :param freq_vmarkers_style: style of the marker in the frequency-domain. If not passed, vertical lines are used.
    :param ylabel: Label to put in the vertical axis. If empty, the Amplitude and unit of it are used.
    :param y_unit_to: Specify the units in which the vertical axis will be displayed
    :param x_unit_to: Specify the units in which the horizontal axis will be displayed
    :param fontsize: the font size
    :return:
    """
    df = copy.copy(dataframe)
    df.reset_index(inplace=True, drop=True)
    _rows_and_cols = []
    # n_rows = 1
    # n_cols = 1
    row_conditions = np.array([1])
    col_conditions = np.array([1])
    if rows_by is not None:
        _rows_and_cols.append(rows_by)
        row_conditions = df[rows_by].astype("category").cat.categories
        # row_domains = df.iloc[df[rows_by].index.values]['domain']
        n_rows = row_conditions.size
    else:
        df.loc[:, 'dummy_row'] = 1
        _rows_and_cols.append('dummy_row')
        # row_domains = df['domain']
        n_rows = 1
    idx_rows = np.arange(row_conditions.size)
    if cols_by is not None:
        _rows_and_cols.append(cols_by)
        col_conditions = df[cols_by].astype("category").cat.categories
        # col_domains = df.iloc[df[cols_by].index.values]['domain']
        n_cols = col_conditions.size
    else:
        df.loc[:, 'dummy_col'] = 1
        _rows_and_cols.append('dummy_col')
        n_cols = 1
        # col_domains = df['domain']
    idx_cols = np.arange(col_conditions.size)

    common_x_axis = np.unique(col_conditions).size == 1
    groups = df.groupby(_rows_and_cols)
    fig_out, ax = plt.subplots(n_rows, n_cols)
    gs = gridspec.GridSpec(n_rows, n_cols)
    for _id, ((_current_row_group, _current_col_group), _group) in enumerate(groups):
        _current_group = _group
        if blend_subcategories:
            _current_group = _group.iloc[[0]].copy().reset_index()
            if _current_group['y'].shape[0] > 1:
                _current_group['y'] = _current_group['y'].apply(lambda x: np.vstack(_group.y).T)
        _idx_row = idx_rows[_current_row_group == row_conditions].squeeze()
        _idx_col = idx_cols[_current_col_group == col_conditions].squeeze()
        _columns_legends = []
        _colors = plt.cm.get_cmap('viridis', _current_group.shape[0])
        _y_unit_to = y_unit_to
        _x_unit_to = x_unit_to
        for _col in _current_group:
            if _col in ['x', 'y']:
                continue
            if len(_current_group[_col].unique()) > 1:
                _columns_legends.append(_col)

        for _i, (_, _row) in enumerate(_current_group.iterrows()):
            _group_label = []
            for _ic, _col in enumerate(_columns_legends):
                _group_label.append(str(_row[_col]))
            _group_label = '/'.join(_group_label)
            y_unit = u.Quantity(1, _row['y_unit'])
            x_unit = u.Quantity(1, _row['x_unit'])
            if _y_unit_to is None:
                _y_unit_to = y_unit.unit
            if _x_unit_to is None:
                _x_unit_to = x_unit.unit
            y = (_row['y'] * y_unit).to(_y_unit_to)
            x = (_row['x'] * x_unit).to(_x_unit_to)
            fs = _row['x_fs']

            _domain = _row['domain']
            if y.ndim == 1:
                y = y.reshape([-1, 1])
            y_single_responses = y

            if _domain == 'time' and sub_average_time_buffer_size is not None:
                fs = 1 / np.mean(np.diff(x))
                used_samples = int(np.floor(y.shape[0] // sub_average_time_buffer_size) * sub_average_time_buffer_size)
                y_f = y[0: used_samples, :]
                y_f = np.transpose(np.reshape(y_f, (sub_average_time_buffer_size, -1, y_f.shape[1]), order='F'),
                                   [0, 2, 1])
                # demean
                y_f = np.mean(y_f, axis=2)
                x = np.arange(0, sub_average_time_buffer_size) / fs
                y_single_responses = y_f - np.mean(y_f, axis=0)

            ax = plt.subplot(gs[_idx_row, _idx_col])
            title = ''
            if title_by == 'row':
                title = '{:}'.format(_current_row_group)

            if title_by == 'col':
                title = '{:}'.format(_current_col_group)

            if title_by == 'both':
                title = '{:} / {:}'.format(_current_row_group, _current_col_group)

            if title_by == 'col':
                if _idx_row == 0:
                    ax.set_title(title, y=1 + title_v_offset, size=fontsize)
            else:
                ax.set_title(title, y=1 + title_v_offset, size=fontsize)

            if _domain == 'time':
                y_mean = np.mean(y_single_responses, axis=1)
                y_sd = np.std(y_single_responses, axis=1)
                y_sem = y_sd / np.sqrt(y_single_responses.shape[1])
                if show_individual_waveforms:
                    ax.plot(x, y_single_responses,
                            linewidth=0.5,
                            alpha=individual_waveforms_alpha,
                            color=_colors(_i)
                            )
            if _domain == 'frequency':
                y_mean = np.mean(np.abs(y_single_responses), axis=1)
                y_sd = np.std(np.abs(y_single_responses), axis=1)
                y_sem = y_sd / np.sqrt(y_single_responses.shape[1])
                if show_individual_waveforms:
                    ax.plot(x, np.abs(y_single_responses),
                            linewidth=0.5,
                            alpha=individual_waveforms_alpha,
                            color=_colors(_i))

            ax.plot(x, y_mean,
                    color=_colors(_i),
                    linewidth=0.8,
                    label=_group_label)
            if ylabel is None:
                ax.set_ylabel('Amplitude [{:}]'.format(y_mean.unit), size=fontsize)
            if show_sem:
                ax.fill_between(x.value,
                                (y_mean - y_sem).value,
                                (y_mean + y_sem).value,
                                alpha=sem_alpha,
                                edgecolor="none",
                                facecolor=_colors(_i))
            if show_sd:
                ax.fill_between(x.value,
                                (y_mean - y_sd).value,
                                (y_mean + y_sd).value,
                                alpha=sd_alpha,
                                edgecolor="none",
                                facecolor=_colors(_i))
            if _idx_col == n_cols - 1:
                ax_row_label = ax.twinx()
                ax_row_label.set_ylabel(_current_row_group, size=fontsize)
                ax_row_label.set_yticklabels([])
                ax_row_label.tick_params(
                    axis='y',  # changes apply to the x-axis
                    which='both',  # both major and minor ticks are affected
                    right=False,  # ticks along the bottom edge are off
                    left=False,  # ticks along the top edge are off
                    labelleft=True)

            if _domain == 'time':
                if time_xlim is not None:
                    ax.set_xlim(time_xlim)
                if time_ylim is not None:
                    ax.set_ylim(time_ylim)
                if time_vmarkers is not None:
                    [ax.axvline(_t, color='k', linestyle=':', linewidth=0.5) for _t in time_vmarkers]
                ax.set_xlabel('Time [{:}]'.format(x_unit.unit), size=fontsize)
            if _domain == 'frequency':
                if freq_xlim is not None:
                    ax.set_xlim(freq_xlim)
                if freq_ylim is not None:
                    ax.set_ylim(freq_ylim)
                if freq_vmarkers is not None:
                    if freq_vmarkers_style is None:
                        [ax.axvline(_f, color='k', linestyle=':', linewidth=0.5) for _f in freq_vmarkers]
                    else:
                        y_min, y_max = ax.get_ylim()
                        [ax.plot(_f, y_max * 0.93,
                                 color='g',
                                 marker=freq_vmarkers_style,
                                 markersize=3) for _f in freq_vmarkers]
                ax.set_xlabel('Frequency [{:}]'.format(x.unit), size=fontsize)
            handles, labels = ax.get_legend_handles_labels()
            if len(labels) > 1 and show_legend:
                fig_out.legend(handles,
                               labels,
                               loc='upper center',
                               fontsize=8,
                               frameon=False,
                               ncol=len(labels))
    if ylabel is None:
        if y_unit_to is not None:
            fig_out.supylabel('Amplitude [{:}]'.format(y_unit_to), size=fontsize)
    else:
        fig_out.supylabel(ylabel, size=fontsize)
    if common_x_axis and _domain == 'time' and x_unit_to is not None:
        fig_out.supxlabel('Time [{:}]'.format(x_unit_to), size=fontsize)
    if common_x_axis and _domain == 'frequency' and x_unit_to is not None:
        fig_out.supxlabel('Frequency [{:}]'.format(x_unit_to), size=fontsize)

    all_axes = fig_out.get_axes()
    for ax in all_axes:
        ax.spines['top'].set_visible(False)
        if not ax.get_subplotspec().is_last_row():  # and row_domains.unique().size == 1:
            ax.set_xticklabels([])
            ax.set_xlabel('')
        ax.spines['left'].set_visible(True)
        ax.spines['right'].set_visible(False)
        ax.tick_params(labelsize=6)
    inch = 2.54
    fig_out.set_size_inches(12.0 / inch, 2.25 * len(row_conditions) / inch)
    fig_out.subplots_adjust(top=0.98, bottom=0.08, hspace=0.0, left=0.15, right=0.95)
    return fig_out


def plot_topographic_maps(dataframe: pd.DataFrame = None,
                          rows_by: str = None,
                          cols_by: str = None,
                          subject_id_column: str = 'subject_id',
                          normalize_by: str = 'subject_id',
                          channels_column: str = 'channel',
                          title: str = '',
                          topographic_value: str = None,
                          layout: str = None,
                          title_by: str = 'row',
                          title_v_offset: float = 0.0,
                          fontsize: float = 6,
                          grid_size: np.complex = 600j,
                          color_map_label: str = None,
                          normalize: bool = False,
                          show_sensors: bool = True,
                          show_sensor_label: bool = False,
                          n_contour_levels: int = 35,
                          max_topographic_value: float = None,
                          min_topographic_value: float = None,
                          contour_line_width: float = 0.1,
                          apply_function: object = lambda x: np.mean(x, axis=2),
                          fun_args: dict = {}
                          ) -> plt.figure:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data)
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy
    :param rows_by: name of the factor in the dataframe for which the rows in the plot will be split
    :param cols_by: name of the factor in the dataframe for which the columns in the plot will be split
    :param subject_id_column: string indicating the column name with subject ids
    :param normalize_by: string indicating the column that will be used to normalized the data
    :param channels_column: name of column containing channel labels
    :param title: title of the figure
    :param topographic_value: name of column containing the value to be shown by the topographic map
    :param layout: path or name of the layout to be used
    :param title_by: string specifying from which factor you want to show the titles in each panel. This can be: "row",
    "col", or "both"
    :param title_v_offset: float specifying the vertical offset of the title
    :param fontsize: the fontsize
    :param grid_size: complex number indicating the size of the grid,
    :param color_map_label: string with the label that would be use of the colourmap. If empty, the default value will
    be the topographic_value
    :param normalize: if True, topographic maps will be normalized within normalize_by
    :param show_sensors: if True, the position of the sensors will be shown
    :param show_sensor_label: if True, the label of the sensor will be shown
    :param n_contour_levels: number of contour lines to plot
    :param min_topographic_value: if given, the colour scale will be restricted to have this value as maximum, otherwise
    it will be determined from the data.
    :param max_topographic_value: if given, the colour scale will be restricted to have this value as minimum, otherwise
    it will be determined from the data.
    :param contour_line_width: float indicating the with (in points) of the contour lines
    :param apply_function: function to be applied to each subgroup
    :param fun_args: extra arguments to be used by apply_function. This is a dictionary with the desired parameters
    :return:
    """
    df = copy.copy(dataframe)
    _rows_and_cols = []
    row_conditions = np.array([1])
    col_conditions = np.array([1])
    if rows_by is not None:
        _rows_and_cols.append(rows_by)
        row_conditions = df[rows_by].astype("category").cat.categories
        n_rows = row_conditions.size
    else:
        df.loc[:, 'dummy_row'] = 1
        _rows_and_cols.append('dummy_row')
        n_rows = 1
    idx_rows = np.arange(row_conditions.size)

    if cols_by is not None:
        _rows_and_cols.append(cols_by)
        col_conditions = df[cols_by].astype("category").cat.categories
        n_cols = col_conditions.size
    else:
        df.loc[:, 'dummy_col'] = 1
        _rows_and_cols.append('dummy_col')
        n_cols = 1
    idx_cols = np.arange(col_conditions.size)

    if not len(_rows_and_cols):
        df['dummy_row'] = 1
        df['dummy_col'] = 1
        _rows_and_cols = ['dummy_row', 'dummy_col']
    # check if input dataframe already contains the potentials. If not, then we extract them
    if 'potentials' in df.keys():
        if normalize:
            if normalize_by is not None:
                df = df.assign(
                    potentials=df.groupby(
                        normalize_by)['potentials'].transform(
                        lambda x: x / np.ma.max(np.abs(np.ma.dstack([_val for _val in x])))))
            else:
                df = df.assign(
                    potentials=df.groupby(
                        [True] * len(df))['potentials'].transform(
                        lambda x: x / np.ma.max(np.abs(np.ma.dstack([_val for _val in x])))))
        sub_groups = df
    else:
        df['topo_value'] = df[topographic_value]
        if normalize:
            df = df.assign(
                topo_value=df.groupby(
                    [normalize_by])[topographic_value].transform(lambda x: x / np.max(np.abs(x))))
        sub_groups = get_topographic_maps_as_df(dataframe=df,
                                                subject_id_column=subject_id_column,
                                                group_by=_rows_and_cols,
                                                channels_column='channel',
                                                topographic_value='topo_value',
                                                layout=layout,
                                                apply_function=apply_function,
                                                fun_args=fun_args
                                                )

    if color_map_label is None:
        color_map_label = topographic_value

    groups = sub_groups.groupby(_rows_and_cols)
    fig_out = plt.figure(constrained_layout=True)
    widths = [1.0] * n_cols
    heights = [1.0] * n_rows
    heights.append(0.1)
    gs = fig_out.add_gridspec(ncols=n_cols, nrows=n_rows + 1,
                              width_ratios=widths,
                              height_ratios=heights)
    if max_topographic_value is None:
        max_sub = sub_groups.potentials.apply(lambda x: np.ma.dstack([_val for _val in x]).max()).max()
    else:
        max_sub = max_topographic_value

    if min_topographic_value is None:
        min_sub = sub_groups.potentials.apply(lambda x: np.ma.dstack([_val for _val in x]).min()).min()
    else:
        min_sub = min_topographic_value
    max_distance = sub_groups.max_distance.max()

    for _id, ((_current_row_group, _current_col_group), _group) in enumerate(groups):
        if _group.shape[0] == 0 or np.all(np.isnan(_group.potentials.values[0])):
            continue
        _idx_row = idx_rows[_current_row_group == row_conditions].squeeze()
        _idx_col = idx_cols[_current_col_group == col_conditions].squeeze()
        ax = plt.subplot(gs[_idx_row, _idx_col])
        panel_title = ''
        if title_by == 'row':
            panel_title = '{:}'.format(_current_row_group)
        if title_by == 'col':
            panel_title = '{:}'.format(_current_col_group)

        if title_by == 'both':
            panel_title = '{:} / {:}'.format(_current_row_group, _current_col_group)

        ax.set_title(panel_title, y=1 + title_v_offset, size=fontsize)
        # std_average = np.ma.mean(_group.potentials.values[0], axis=2)
        if _group.potentials.values.size > 1:
            std_average = _group.potentials.apply(
                lambda x: apply_function(np.ma.dstack([_val for _val in x]), *fun_args)).values[0]
        else:
            std_average = _group.potentials.values[0]
        ax_im = ax.imshow(std_average.T, origin='lower',
                          extent=(-max_distance, max_distance, -max_distance, max_distance),
                          vmin=min_sub,
                          vmax=max_sub,
                          aspect=1.0)
        ax_im.set_cmap('nipy_spectral')

        levels = np.linspace(min_sub, max_sub, n_contour_levels)
        ax.contour(std_average.T,
                   levels,
                   origin='lower',
                   extent=(-max_distance, max_distance, -max_distance, max_distance),
                   linewidths=contour_line_width,
                   colors='k',
                   linestyles='solid')
        ax.autoscale(enable=False)
        # plot color bar

        if _id == 0:
            c_bar_ax = plt.subplot(gs[-1, :])
            c_bar = fig_out.colorbar(ax_im, cax=c_bar_ax, orientation='horizontal', format='%.1f')
            c_bar.set_label(color_map_label, fontsize=fontsize)
            tick_locator = ticker.MaxNLocator(nbins=3)
            c_bar.locator = tick_locator
            c_bar.update_ticks()
            c_bar.ax.tick_params(labelsize=fontsize)
        channels = _group.channels.values[0]
        channel_labels = [ch['label'] for ch in channels]
        ax.plot(0, max_distance * 1.0, '|', markersize=8, color='k')
        if show_sensors:
            _lay = Layout()
            _layout = _lay.get_layout(file_name=layout)
            _layout = np.array([_l for _l in _layout if _l.label not in ['COMNT', 'SCALE']])
            for i, lay in enumerate(_layout):
                if lay.label in channel_labels:
                    ax.plot(lay.x, lay.y, 'o', color='b', markersize=0.2)
                else:
                    ax.plot(lay.x, lay.y, 'o', color='grey', markersize=0.2)
                if show_sensor_label:
                    ax.text(lay.x, lay.y, s=lay.label, fontsize=6)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis('off')

    all_axes = fig_out.get_axes()
    for ax in all_axes:
        if ax == c_bar.ax:
            continue
        ax.spines['top'].set_visible(False)
        ax.set_xticklabels([])
        ax.set_xlabel('')
        ax.spines['left'].set_visible(True)
        ax.spines['right'].set_visible(False)
    inch = 2.54
    fig_out.suptitle(title)
    fig_out.set_size_inches(3.2 * len(col_conditions) / inch, h=3.2 * len(row_conditions) / inch)
    return fig_out


def get_topographic_maps(df,
                         subject_id_column: str = 'subject_id',
                         channels_column: str = 'channel',
                         topographic_value: str = None,
                         layout: str = None,
                         grid_size: np.complex = 600j,
                         apply_function: object = None,
                         fun_args: dict = {}):
    _lay = Layout()
    _layout = _lay.get_layout(file_name=layout)
    _layout = np.array([_l for _l in _layout if _l.label not in ['COMNT', 'SCALE']])
    _single_responses = np.array([])

    _subject_groups = df.groupby(subject_id_column)
    for _, (_id_sub, _sub_group) in enumerate(_subject_groups):
        x = []
        y = []
        z = []
        channels = []
        max_potential = -np.inf
        for _i, (_, _row) in enumerate(_sub_group.iterrows()):
            ch = _row[channels_column]
            amp = _row[topographic_value]
            for _l in _layout:
                if _l.label == ch:
                    x.append(_l.x)
                    y.append(_l.y)
                    z.append(amp)
            channels.append({'label': ch})

        _potentials, max_distance = eegpt.interpolate_potential_fields(x=np.array(x).reshape(-1, 1),
                                                                       y=np.array(y).reshape(-1, 1),
                                                                       z=np.array(z).reshape(-1, 1),
                                                                       grid=grid_size)
        max_potential
        if not _single_responses.size:
            _single_responses = _potentials[:, :, None]
        else:
            _single_responses = np.ma.dstack((_single_responses, _potentials))
    if apply_function is not None:
        _single_responses = apply_function(_single_responses, **fun_args)
    out = pd.Series(data={'potentials': _single_responses,
                          'max_distance': max_distance,
                          'channels': channels})
    return out


def get_topographic_maps_as_df(dataframe: pd.DataFrame = None,
                               group_by: [str] = None,
                               subject_id_column: str = 'subject_id',
                               channels_column: str = 'channel',
                               topographic_value: str = None,
                               layout: str = None,
                               grid_size: np.complex = 600j,
                               apply_function: object = np.mean,
                               fun_args: dict = {}
                               ) -> pd.DataFrame:
    """
    This function will plot the waveforms contained in a pandas dataframe read using the sqlite_waveforms_to_pandas
    function of pEEGy.
    The rows and columns of the output plot are specified by the factors of the dataframe.
    The output will show the data for each of those factors (both individual and average data).
    :param dataframe: a pandas dataframe returned by sqlite_waveforms_to_pandas function of pEEGy.
    :param group_by: name of factors to group and obtain potentials.
    :param subject_id_column: string indicating the column name with subject ids
    :param channels_column: name of column containing channel labels
    :param topographic_value: name of column containing the value to be shown by the topographic map
    :param layout: path or name of the layout to be used
    :param grid_size: complex number indicating the size of the grid,
    :return:
    """
    df = copy.copy(dataframe)
    df['topo_value'] = df[topographic_value]
    group_by = [True] * len(df) if group_by is None else group_by
    sub_groups = df.groupby(group_by, group_keys=False).apply(
        lambda x: get_topographic_maps(x,
                                       subject_id_column=subject_id_column,
                                       layout=layout,
                                       channels_column=channels_column,
                                       topographic_value='topo_value',
                                       grid_size=grid_size,
                                       apply_function=apply_function,
                                       fun_args=fun_args)).reset_index()
    return sub_groups



import copy

import numpy as np
import astropy.units as u
import pandas as pd


class SingleEvent(object):
    def __init__(self,
                 code: float = None,
                 time_pos: u.quantity.Quantity = None,
                 dur: u.quantity.Quantity = None,
                 bad_event: bool = False):
        """
        Class for a single (trigger) event, which is characterized by having a numerical identifier (code), an
        associated time point [s], and a duration [s].
        Parameters
        ----------
        code: Trigger event code (usually some number)
        time_pos: Time in seconds of trigger event
        dur: Duration of trigger event in seconds
        bad_event: indicate if this event is considered bad
        """
        self.code = code
        self.time_pos = time_pos
        self.dur = dur
        self.bad_event = bad_event

    def __repr__(self):
        return (f'{self.__class__.__name__}('f'{self.code!r}, {self.time_pos!r}, {self.dur!r}')

    def __str__(self):
        return 'Single event at time {} s with code {} and duration {} s'.format(self.time_pos, self.code, self.dur)


class Events(object):
    def __init__(self, events: np.array([SingleEvent]) = None):
        """
        Defines event class
        :param events: raw events array
        """
        self._events = events
        self._mask = None

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value: float = None):
        self._mask = value
        print('Event mask {:} set'.format(self.mask))
        print(self.summary().to_string(index=False))

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'{self.events.shape!r}')

    def __str__(self):
        return 'Event array of size {}'.format(self.events.shape)

    def events(self, include_bad_events: bool = False):
        out = []
        if self.mask is not None:
            for _ev in self._events:
                new_event = copy.copy(_ev)
                new_event.code = float(int(new_event.code) & int(self.mask))
                out.append(new_event)
            out = np.array(out)
        else:
            if include_bad_events:
                out = self._events
            else:
                out = np.array([_ev for _ev in self._events if not _ev.bad_event])
        return out

    def get_events(self, code=None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev for _ev in self.events(include_bad_events=include_bad_events)])
        else:
            return np.array([_ev for _ev in self.events(include_bad_events=include_bad_events) if _ev.code == code])

    def get_events_index(self, code=None, fs=None, include_bad_events: bool = False):
        if code is None:
            return np.array([int(_ev.time_pos * fs) for _ev in self.events(include_bad_events=include_bad_events)])
        else:
            return np.array([int(_ev.time_pos * fs) for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code == code])

    def get_events_time(self, code=None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.time_pos.to(u.s).value for
                             _ev in self.events(include_bad_events=include_bad_events)]) * u.s
        else:
            return np.array([_ev.time_pos.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code == code]) * u.s

    def get_events_duration(self, code=None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.dur.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)]) * u.s
        else:
            return np.array([_ev.dur.to(u.s).value for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code == code]) * u.s

    def get_events_code(self, code=None, include_bad_events: bool = False):
        if code is None:
            return np.array([_ev.code for _ev in self.events(include_bad_events=include_bad_events)])
        else:
            return np.array([_ev.code for _ev in self.events(include_bad_events=include_bad_events)
                             if _ev.code == code])

    def set_events_status(self,
                          event_idx: np.array,
                          bad_event: bool = False):
        for _ev in self._events[event_idx]:
            _ev.bad_event = bad_event

    def interpolate_events(self, scaling_factor=1.0, code=None, events_mask=None):
        """
        This function interpolates time events between consecutive event. The time position of new events is defined
        by the scaling factor. For example, an scaling_factor of 2.0 will add a new trigger event exactly in the middle
        time between two events
        :param scaling_factor: a float that determine the number of new events that will be generated
        :param code: event code to be interpolated
        :param events_mask: integer value used to masker triggers codes. This is useful to ignore triggers inputs avove
        a particular value. For example, if only the first 8 trigger inputs were used (max decimal value is 255), in a
        system with 16 trigger inputs, then the masker could be set to 255 to ignore any trigger from trigger inputs 9
        to 16.
        :return: new an Events class containing the new interpolated events
        """
        self.mask = events_mask
        _times = self.get_events_time(code=code)
        _durations = self.get_events_duration(code=code)
        _codes = self.get_events_code(code=code)
        _times_intervals = np.diff(_times)
        _times_intervals = np.append(_times_intervals, np.mean(_times_intervals))
        _new_events = []
        for _ini_time, _length, _dur, _code in zip(_times, _times_intervals, _durations, _codes):
            _new_times = np.arange(_ini_time.to(u.s).value,
                                   (_ini_time + _length).to(u.s).value,
                                   (_length / scaling_factor).to(u.s).value) * u.s
            [_new_events.append(SingleEvent(time_pos=_t, dur=_dur, code=_code)) for _t in _new_times]
        return Events(events=np.array(_new_events))

    def summary(self):
        out = pd.DataFrame()
        for i, _code in enumerate(np.unique(self.get_events_code())):
            out = pd.concat([out,
                             pd.DataFrame.from_dict(
                                 {'event_code': [_code],
                                  'n_events': [self.get_events_code(code=_code).size]})],
                            ignore_index=True)
        return out

    def clip(self, ini_time: u.Quantity = 0 * u.s, end_time: u.Quantity = 0 * u.s):
        self._events = np.array([_ev for _ev in self._events if _ev.time_pos >= ini_time])
        self._events = np.array([_ev for _ev in self._events if _ev.time_pos <= end_time])

    def set_offset(self, time_offset: u.Quantity = 0 * u.s):
        self._events = np.array([
            SingleEvent(time_pos=_ev.time_pos - time_offset,
                        dur=_ev.dur,
                        code=_ev.code) for _ev in self._events])



import numpy as np
from peegy.processing.tools.filters.resampling import eeg_resampling
import pyfftw
import multiprocessing
import copy
from peegy.definitions.channel_definitions import ElectrodeType
from peegy.processing.statistics.definitions import HotellingTSquareFrequencyTest
from peegy.processing.tools.detection.definitions import TimePeakWindow, EegPeak, PeakToPeakMeasure, TimeROI
__author__ = 'jundurraga'


class EegChannel(object):
    def __init__(self, **kwargs):
        self.number = kwargs.get('number', None)
        self.label = kwargs.get('label', str(self.number))
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)
        self.h = kwargs.get('h', None)
        self.type = kwargs.get('type', ElectrodeType.scalp)
        self.valid = kwargs.get('valid', True)
        # if fixed_polarity is true the polarity of this channel will not be reversed when required
        # by class EegAverageEpochs
        self.fixed_polarity = kwargs.get('fixed_polarity', False)


class EegAverageEpochs(object):
    def __init__(self, **kwargs):
        self.invert_polarity = kwargs.get('invert_polarity', False)
        self.channels = kwargs.get('channels', None)
        self.rn = kwargs.get('rn', None)
        self.cum_rn = kwargs.get('cum_rn', None)
        self.snr = kwargs.get('snr', None)
        self.cum_snr = kwargs.get('cum_snr', None)
        self.n_samples_block = kwargs.get('n_samples_block', None)
        self.signal_variance = kwargs.get('signal_variance', None)
        self.amplitude_units = kwargs.get('amplitude_units', None)
        self.time_units = kwargs.get('time_units', None)
        self.frequency_bins = kwargs.get('frequency_bins', None)
        self.frequency_test = kwargs.get('frequency_test', [HotellingTSquareFrequencyTest])
        self.data_processing = kwargs.get('data_processing', {})
        self.peak_times = kwargs.get('time_peaks', [EegPeak()])
        self.peak_frequency = kwargs.get('peak_frequency', [EegPeak()])
        self.peak_amplitudes = kwargs.get('peak_amplitudes', [PeakToPeakMeasure()])
        self.peak_time_windows = kwargs.get('peak_time_windows', [TimePeakWindow()])
        self.resampling_factor = 1.0
        self.descriptor = kwargs.get('descriptor', '')
        self.roi_windows = kwargs.get('roi_windows', np.array([TimeROI()]))
        self._average = kwargs.get('average', np.array([]))
        self._fs = kwargs.get('fs', None)
        self._rfft_average = kwargs.get('rfft_average', np.array([]))

    def set_average(self, value=np.array([])):
        self._average = value

    def get_average(self):
        if self.channels is None:
            return np.array([])
        else:
            return self._average * (-1.0) ** (
                self.invert_polarity & ~np.array([_ch.fixed_polarity for _ch in self.channels])).astype(int)

    average = property(get_average, set_average)

    def set_rfft_average(self):
        fft = pyfftw.builders.rfft(self._average, overwrite_input=False, planner_effort='FFTW_ESTIMATE', axis=0,
                                   threads=multiprocessing.cpu_count())
        _fft = np.abs(fft())
        # scale data
        _fft /= _fft.shape[0]
        self._rfft_average = _fft

    def get_rfft_average(self):
        return self._rfft_average

    rfft_average = property(get_rfft_average)

    def get_fs(self):
        return self._fs

    def set_fs(self, new_value):
        self._fs = new_value

    fs = property(get_fs, set_fs)

    def get_time(self):
        return np.arange(self._average.shape[0]) / self._fs

    time = property(get_time)

    def get_frequency(self):
        return np.arange(self._rfft_average.shape[0]) * self._fs / self._average.shape[0]

    frequency = property(get_frequency)

    def time_to_samples(self, in_time):
        return np.rint(in_time * self._fs).astype(np.int64)

    def frequency_to_samples(self, in_freq):
        return np.rint(in_freq / (self._fs / 2.0) * self.frequency.size).astype(np.int64)

    def append_eeg_ave_channel(self, new_channel_data=np.array([]),
                               rn=None,
                               snr=None,
                               signal_variance=None,
                               label=None,
                               electrode_type=ElectrodeType.artificial,
                               **kwargs):
        self._average = np.hstack((self._average, new_channel_data))
        self.rn = np.append(self.rn, rn)
        for _i, _ in enumerate(self.snr):
            self.snr[_i] = np.append(self.snr[_i], snr[_i])
        for _i, _ in enumerate(self.signal_variance):
            self.signal_variance[_i] = np.append(self.signal_variance[_i], signal_variance[_i])
        new_channel = EegChannel(label=label, number=len(self.channels) + 1, type=type, **kwargs)
        self.channels = np.append(self.channels, new_channel)

    def resampling(self, new_fs=None):
        self._average, _factor = eeg_resampling(x=self._average, new_fs=new_fs, fs=self._fs)
        self.resampling_factor = _factor
        self._fs *= _factor
        self.set_rfft_average()
        # scale frequency tested to new sampling rate
        # if self.frequency_test is not None:
        #     for _h_t in self.frequency_test:
        #         _h_t.frequency_tested = self.frequency[_h_t.frequency_bin]

    def is_scalp_channel(self):
        return np.array([_ch.type == ElectrodeType.scalp for _ch in self.channels])

    def get_roi_measures(self, roi_windows=np.array([TimeROI()])):
        out = []
        for _roi in roi_windows:
            _samples = self.time_to_samples(_roi.interval)
            for _i, _channel in enumerate(self.channels):
                _roi.channel = _channel.label
                _data = self._average[_samples[0]: _samples[1], _i]

                if _roi.measure == 'snr':
                    s_var = np.var(_data, ddof=1)
                    if self.rn[_i] == 0:
                        snr = np.inf
                    else:
                        snr = s_var / self.rn[_i] ** 2.0 - 1.0
                    _roi.value = snr
                if _roi.measure == 'rms':
                    _roi.value = np.sqrt(np.mean(np.square(_data)))
                out.append(copy.copy(_roi))
        return out

    def get_unique_interval_rois(self, type='snr'):
        out = []
        if self.roi_windows.size:
            intervals = np.array([_roi.interval for _roi in self.roi_windows if _roi.label == type])
            b = np.ascontiguousarray(intervals).view(np.dtype((np.void, intervals.dtype.itemsize * intervals.shape[1])))
            _, idx = np.unique(b, return_index=True)
            out = self.roi_windows[idx]
        return out

    def get_rois_by_type(self, type='snr'):
        out = []
        if self.roi_windows is not None:
            out = np.array([_roi for _roi in self.roi_windows if _roi.measure == type])
        return out

    def roi_windows_in_samples(self):
        roi_samples = []
        for _i, _roi in enumerate(self.get_unique_interval_rois()):
            roi_samples.append(self.time_to_samples(_roi.interval))
        return roi_samples

    def get_roi_snr_and_rn(self):
        _roi_windows = self.get_rois_by_type(type='snr')
        _cum_snr = self.cum_snr
        _cum_rn = self.cum_rn
        _n_trials = np.cumsum(self.n_samples_block)
        _ch_idx = self.is_scalp_channel()
        _channels = self.channels[_ch_idx]
        out = []
        for _n, _roi_snrs in zip(_n_trials, _cum_snr):
            for _roi, _all_ch_snrs in zip(_roi_windows, _roi_snrs):
                for _ch, _snr in zip(_channels, _all_ch_snrs):
                    _roi = TimeROI(**{'interval': _roi.interval,
                                      'measure': 'snr',
                                      'condition': _n,
                                      'label': _roi.label,
                                      'channel': _ch.label,
                                      'value': _snr})
                    out.append(_roi)

        for channels_rn, _n in zip(_cum_rn, _n_trials):
            for _ch, _rn in zip(_channels, channels_rn):
                _roi = TimeROI(**{'interval': [],
                                  'measure': 'rn',
                                  'condition': _n,
                                  'label': 'rn',
                                  'channel': _ch.label,
                                  'value': _rn})
                out.append(_roi)
        return out

    def get_max_snr_per_channel(self):
        max_snr = np.nanmax(np.atleast_2d(self.snr), axis=0)
        return np.atleast_1d(np.squeeze(max_snr))

    def get_max_s_var_per_channel(self):
        max_var = np.nanmax(self.signal_variance, axis=0)
        return np.atleast_1d(np.squeeze(max_var))


def signal_information(**kwargs):
    # ensure there are this mandatory fields
    out = {'type': kwargs.pop('type', None),
           'condition': kwargs.pop('condition', None)}
    for _item, _val in kwargs.items():
        out[_item] = _val
    return out



import numpy as np
import multiprocessing
import pyfftw
import astropy.units as u
from tqdm import tqdm
from .. multiprocessing.multiprocessesing_filter import filt_data
from .. filters.eegFiltering import bandpass_fir_win
__author__ = 'jundurraga'


def eeg_resampling(x: np.array = np.array([]),
                   new_fs: u.Quantity = None,
                   fs: u.Quantity = None,
                   transition_width: u.Quantity = 1.0 * u.Hz,
                   blocks=8) -> np.array:
    factor = new_fs / fs
    if factor == 1.0:
        return x, factor

    factor = np.round(factor * x.shape[0]) / x.shape[0]
    updated_nfft = int((x.shape[0] * factor))

    if factor < 1:
        _b = bandpass_fir_win(high_pass=None,
                              low_pass=new_fs / 2 - transition_width,
                              fs=fs,
                              transition_width=transition_width)
        _b = _b * u.dimensionless_unscaled
        # apply antialiasing
        x = filt_data(data=x, b=_b)

    blocks = np.minimum(x.shape[1], blocks)
    xfilt = np.zeros((updated_nfft, x.shape[1]))
    for b in tqdm(np.arange((np.ceil(x.shape[1]) / blocks)), desc='Resampling'):
        pos_ini = int(b * blocks)
        pos_end = int(np.minimum((b + 1) * blocks, x.shape[1]) + 1)
        fft = pyfftw.builders.rfft(x[:, pos_ini:pos_end],
                                   overwrite_input=False,
                                   planner_effort='FFTW_ESTIMATE',
                                   axis=0,
                                   threads=multiprocessing.cpu_count())
        fft_data = fft()
        ifft = pyfftw.builders.irfft(fft_data,
                                     n=updated_nfft,
                                     overwrite_input=False,
                                     planner_effort='FFTW_ESTIMATE',
                                     axis=0,
                                     threads=multiprocessing.cpu_count())
        sub_set = ifft()
        xfilt[:, pos_ini:pos_end] = sub_set
        if pos_end == x.shape[1]:
            break
    return xfilt * factor * x.unit, factor



import matplotlib.pyplot as plt
import os
import graphviz
import PIL as pil
import pydot
import tempfile
from PyQt5.QtCore import QLibraryInfo
from peegy.processing.pipe.definitions import InputOutputProcess
import pyqtgraph as pg
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class PipePool(dict):
    """This class manages InputOutputProcess appended to it.
    When InputOutputProcess are appended, to the PipePool, they can be called by calling the run function of the
    PipePool. If an element has been already run, this won't be called again.
    """

    def __init__(self):
        super(PipePool, self).__init__()

    def __del__(self):
        plt.close('all')
        pg.cleanup()
        pg.QtWidgets.QApplication.processEvents()

    def __setitem__(self, key, value):
        # optional processing here
        assert isinstance(value, InputOutputProcess)
        key = self.ensure_unique_name(label=key)
        super(PipePool, self).__setitem__(key, value)
        # we add the new item as class variable
        setattr(self, key, value)
        # set name of InoutOutputProcess
        value.name = key

    def append(self, item: object, name=None):
        if name is None:
            name = type(item).__name__
        self[name] = item

    def ensure_unique_name(self, label: str = None):
        all_names = [_key for _key in self.keys()]
        _label = label
        count = 0
        while _label in all_names:
            _label = label + '_' + str(count)
            count = count + 1
        if count > 0:
            print('PipePool item "{:}" already exists. Renamed to "{:}"'.format(label, _label))
        return _label

    def __getitem__(self, key):
        out = None
        if key in self.keys():
            out = super(PipePool, self).__getitem__(key)
        return out

    def run(self, force: bool = False):
        for _item, _value in self.items():
            if _value.ready and not force:
                continue
            _value.run()
            _value.ready = True

    def diagram(self,
                file_name: str = '',
                return_figure=False,
                dpi=300,
                ):
        """
        This function generates a flow diagram with the available connections in the pipeline
        :param file_name: string indicating the file name to save pipeline diagram
        :param return_figure: boolean indicating if figure should be returned
        :param dpi: resolution of output figure

        """
        _file, file_extension = os.path.splitext(file_name)
        _temp_file = tempfile.mkdtemp() + os.sep + 'diagram'
        gviz = graphviz.Digraph("Pipeline",
                                filename=_temp_file,
                                node_attr={'color': 'lightblue2', 'style': 'filled'},
                                format=file_extension.replace('.', ''))
        gviz.attr(rankdir='TB', size='7,14')
        items = list(self.items())
        for _idx_1 in list(range(len(self))):
            _current_list = []
            _assigned = False
            for _idx_2 in list(range(len(self))):
                if _idx_1 == _idx_2:
                    continue
                (_name_1, process_1) = items[_idx_1]
                (_name_2, process_2) = items[_idx_2]
                if process_1 is not None and process_2 is not None:
                    process_2_list = []
                    if isinstance(process_1, InputOutputProcess):
                        if isinstance(process_2.input_process, InputOutputProcess):
                            process_2_list = [process_2.input_process]
                        elif isinstance(process_2.input_process, list) and all(isinstance(x, InputOutputProcess)
                                                                               for x in process_2.input_process):
                            process_2_list = process_2.input_process
                        else:
                            raise ValueError('Either an InputOutputProcess or list of them should be used')
                    for _p2 in process_2_list:
                        if process_1 == _p2:
                            _current_list.append(_name_2)
                            gviz.edge(_name_1, _name_2)
                            _assigned = True
            if not _assigned:
                gviz.node(_name_1)
        gviz.render(_temp_file)
        # use numpy to construct an array from the bytes
        gviz.save(_temp_file)
        (graph,) = pydot.graph_from_dot_file(_temp_file)
        graph.write_png(_file + '.png', prog=['dot', '-Gdpi={:}'.format(dpi)])
        fig_out = None

        if return_figure:
            img = pil.Image.open(_file + '.png')
            fig_out = plt.figure()
            ax = fig_out.add_subplot(111)
            ax.imshow(img, aspect=1)
            ax.axis('off')
            fig_out.tight_layout()
        return fig_out



from peegy.processing.tools.epochs_processing_tools import et_subtract_oeg_template
from peegy.processing.tools.multiprocessing.multiprocessesing_filter import filt_data
from peegy.processing.tools.template_generator.auditory_waveforms import eog
import numpy as np
from numpy import matlib
import matplotlib.pyplot as plt
import astropy.units as u
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')


def unitary_step(x: np.array = None):
    return (x > 0).astype(np.float)


np.random.seed(7123)
n_ch = 10
fs = 16384.0 * u.Hz
recording_time = 180
eog_rate = 1 / 4 * u.Hz
time = np.arange(0, recording_time, 1/fs.to(u.Hz).value)
time_artefact = np.arange(0, 1/3, 1/fs.to(u.Hz).value)
n_events = int(recording_time // (1 / eog_rate.to(u.Hz).value))
# we generate a fake low frequency artefact signal and generate a mixing matrix
eog1_template, _ = eog(fs)
# eog1_template *= unitary_step(eog1_template)

events = np.zeros(time.size, )
_intervals = 1/eog_rate.to(u.Hz).value + 0.5 * np.random.randn(n_events, )
event_times = np.cumsum(_intervals)
event_times = event_times[event_times < recording_time]
_idx = np.minimum(event_times * fs.to(u.Hz).value, time.size - 1).astype(int)
events[_idx] = 1 + 0.5 * np.random.rand(_idx.size)
eog1 = filt_data(eog1_template.reshape(-1, 1), events.flatten(), mode='full').flatten()
eog1 = eog1[0: time.size]
w1 = np.random.rand(1, n_ch)
eog1mix = matlib.repmat(eog1, n_ch, 1).T * w1
eog1mix[:, -1] = eog1

# generate EEG data at a higher frequency
ep = 0.1 * np.sin(2 * np.pi * 137.0 * time)
w3 = np.random.rand(1, n_ch)
epochs = matlib.repmat(ep, n_ch, 1).T * w3 * u.uV
epochs[:, -1:] = 0

# generate noise and mixing matrix
ns = np.random.randn(*time.shape)
w4 = np.random.rand(1, n_ch)
noise = 40 * matlib.repmat(ns, n_ch, 1).T * w4 * u.uV

# mix all data
data = eog1mix + epochs + noise
ref_idx = np.array([9])

# apply rise-fade window
fade_in_out = np.ones(data.shape)
fade_len = round(time.size // 40)
fade_in_out[0: fade_len, :] = (np.sin(2 * np.pi * (1 / (4 * fade_len / fs.to(u.Hz).value)) *
                                      time[0: fade_len]) ** 2)[:, None]
fade_in_out[-fade_len:, :] = (np.sin(2 * np.pi * (1 / (4 * fade_len / fs.to(u.Hz).value)) *
                                     time[0: fade_len] + np.pi / 2) ** 2)[:, None]
data *= fade_in_out
# remove artefacts
clean_data, fig_results = et_subtract_oeg_template(data=data,
                                                   idx_ref=ref_idx,
                                                   fs=fs,
                                                   plot_results=True,
                                                   high_pass=0.01 * u.Hz,
                                                   low_pass=20 * u.Hz,
                                                   minimum_interval_width=0.2 * u.s,
                                                   kernel_bandwidth=0.15)

# get frequency responses
freq = np.arange(0, time.size) * fs / time.size
yfft_data = np.abs(np.fft.fft(data, axis=0))
yfft_clean = np.abs(np.fft.fft(clean_data, axis=0))

fig = plt.figure()
ax = fig.add_subplot(221)
ax.plot(time, np.squeeze(data[:, 0]), label='contaminated')
ax.plot(time, np.squeeze(clean_data[:,  0]), label='clean')
ax = fig.add_subplot(222)
ax.plot(freq, np.squeeze(yfft_data[:, 0]), label='contaminated fft')
ax.plot(freq, np.squeeze(yfft_clean[:, 0]), label='clean fft')
ax.set_xlim(0, 200)
plt.legend()
plt.show()



"""
.. _tut-FFR-artefact-removal-test-sim:

########################################################
FFR Artefact removal example (Simulated)
########################################################

In this example we simulate an FFR, and we remove artefacts via a regression method.
This method attempts to regress out a reference artefact signal from any other channel by estimating the transmission
index (scalar factor) that is then applied to the reference artefact signal to then subtract the scaled signal from
each channel.

.. contents:: Page contents
   :local:
   :depth: 2
"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
import copy
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels, ReSampling
from peegy.processing.pipe.epochs import EpochData, AverageEpochsFrequencyDomain
from peegy.processing.pipe.regression import RegressOutArtifact
from peegy.processing.pipe.plot import PlotTopographicMap, PlotWaveforms
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
from peegy.processing.tools.filters.resampling import eeg_resampling
import numpy as np

# %%
# Generate some data
# ========================
# We generate some artefactual and frequency-following data. In this example the stimulation source and the brain
# source have different delays


fs = 2048.0 * u.Hz
epoch_length = 1.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
burst_duration = 0.375 * u.s
ffr_frequencies = np.array([60, 120, 180])  * u.Hz
ffr_frequencies = np.ceil(burst_duration * ffr_frequencies) / burst_duration  # fit to burst_duration length
alternating_polarity = False  # stimulus is alternating in polarity every presentation
# here we pick some random frequencies to test statistical detection
random_frequencies = np.unique(np.random.randint(50, 200, 3)) * u.Hz
stim_delay = 0.001 * u.s  # neural delay in secs
brain_delay = 0.0237 * u.s
time = np.arange(0, burst_duration.to(u.s).value, 1 / fs.to(u.Hz).value).reshape(-1, 1) * u.s
# stimulation waveform
_original_stimulus_waveform = np.sum(
    3 * np.sin(2 * np.pi * u.rad * ffr_frequencies * time),
    axis=1).reshape(-1, 1) * 1.0 * u.uV  # generates ffr artifacts with 1 uV amplitude

stimulus_waveform = np.pad(_original_stimulus_waveform, ((int(fs * stim_delay), int(fs * brain_delay)),
                                                         (0, 0)), 'constant', constant_values=(0, 0))
# brain response
template_waveform = np.pad(_original_stimulus_waveform, ((int(fs * stim_delay) + int(fs * brain_delay), 0), (0, 0)),
                           'constant', constant_values=(0, 0))
template_waveform *= 0.2  # 0.2 uV amplitude and a delay
template_waveform[template_waveform < 0] = 0  # rectify

n_channels = 16
event_times = np.arange(0, 360.0, epoch_length.to(u.s).value) * u.s
artefact_mixing_matrix = np.zeros(n_channels)
artefact_mixing_matrix[5:12] = 1
artefact_mixing_matrix = np.diag(artefact_mixing_matrix)
reader = GenerateInputData(template_waveform=template_waveform,
                           mixing_matrix=np.diag(np.arange(n_channels) / n_channels),
                           stimulus_waveform=stimulus_waveform,
                           alternating_polarity=alternating_polarity,
                           artefacts_mixing_matrix=artefact_mixing_matrix,
                           fs=fs,
                           n_channels=n_channels,
                           snr=0.5,
                           layout_file_name='biosemi16.lay',
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='ffr_artifact_test',
                           noise_seed=0)
reader.run()

# %%
# Start the pipeline
# ========================


new_fs = fs / 2
pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])

pipeline['down_sampled'] = ReSampling(pipeline['channel_cleaned'],
                                      new_sampling_rate=new_fs)
pipeline['time_filtered_data'] = FilterData(pipeline['down_sampled'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=None)
pipeline.run()

# replace neural data in reader to generate figures too
target = copy.copy(reader)
target.output_node.data = target.simulated_neural_response
pipeline['target_referenced'] = ReferenceData(reader,
                                              reference_channels=['Cz'],
                                              invert_polarity=True)
pipeline['down_sampled_target'] = ReSampling(pipeline['target_referenced'],
                                             new_sampling_rate=new_fs)

# %%
# Now we down sample the stimulus waveform
# ---------------------------------------------
# Here, the stimulus waveform (usually recorded as an EEG channel) is down sampled to have the same temporal resolution
# as the simulated brain data


rs_stimulus_waveform, _ = eeg_resampling(x=stimulus_waveform,
                                         new_fs=new_fs,
                                         fs=fs)

# %%
# Now regress out the stimulus waveform from the simulated EEG datagit
# --------------------------------------------------------------------


pipeline['artifact_free'] = RegressOutArtifact(pipeline['time_filtered_data'],
                                               event_code=1.0,
                                               alternating_polarity=alternating_polarity,
                                               stimulus_waveform=rs_stimulus_waveform,
                                               method='regression'
                                               )
pipeline.run()

# %%
# Comparing recorded and cleaned waveforms
# ---------------------------------------------

pipeline['artifact_free'].plot(plot_input=True,
                               plot_output=True,
                               ch_to_plot=np.array(['O2', 'T8', 'T7']),
                               interactive=False)

# %%
# Get Epochs
# ---------------------------------------------
# We partition the data into epochs or trials based on the event code used.

pipeline['time_epochs_no_artefacts'] = EpochData(pipeline['artifact_free'],
                                                 event_code=1.0,
                                                 base_line_correction=False,
                                                 post_stimulus_interval=burst_duration / 1.0)

pipeline['time_epochs_with_artefacts'] = EpochData(pipeline['time_filtered_data'],
                                                   event_code=1.0,
                                                   base_line_correction=False,
                                                   post_stimulus_interval=burst_duration / 1.0)

pipeline['time_target'] = EpochData(pipeline['down_sampled_target'],
                                    event_code=1.0,
                                    base_line_correction=False,
                                    post_stimulus_interval=burst_duration / 1.0)
pipeline.run()

# %%
# Get average in frequency-domain of data without using spatial-filtering
# ---------------------------------------------------------------------------


pipeline['fft_ave_no_artefacts'] = AverageEpochsFrequencyDomain(pipeline['time_epochs_no_artefacts'],
                                                                test_frequencies=np.concatenate((
                                                                    ffr_frequencies, random_frequencies)),
                                                                n_fft=int(burst_duration * new_fs)
                                                                )
pipeline.run()
pipeline['fft_ave_with_artefacts'] = AverageEpochsFrequencyDomain(pipeline['time_epochs_with_artefacts'],
                                                                  test_frequencies=np.concatenate((
                                                                      ffr_frequencies, random_frequencies)),
                                                                  n_fft=int(burst_duration * new_fs)
                                                                  )
pipeline.run()
pipeline['fft_target'] = AverageEpochsFrequencyDomain(pipeline['time_target'],
                                                      test_frequencies=np.concatenate((
                                                          ffr_frequencies, random_frequencies)),
                                                      n_fft=int(burst_duration * new_fs)
                                                      )
pipeline.run()


# %%
# Plot frequency-domain average of data
# ---------------------------------------------------------------------------


pipeline['plotter_1'] = PlotTopographicMap(pipeline['fft_ave_no_artefacts'],
                                           topographic_channels=np.array(['O2', 'T8', 'T7']),
                                           plot_x_lim=[50, 200],
                                           plot_y_lim=[0, 0.5],
                                           return_figures=True,
                                           user_naming_rule='fft_ave_no_artefacts')

pipeline['plotter_2'] = PlotTopographicMap(pipeline['fft_ave_with_artefacts'],
                                           topographic_channels=np.array(['O2', 'T8', 'T7']),
                                           plot_x_lim=[50, 200],
                                           plot_y_lim=[0, 0.5],
                                           return_figures=True,
                                           user_naming_rule='fft_ave_with_artefacts')

pipeline['plotter_3'] = PlotTopographicMap(pipeline['fft_target'],
                                           topographic_channels=np.array(['O2', 'T8', 'T7']),
                                           plot_x_lim=[50, 200],
                                           plot_y_lim=[0, 0.5],
                                           return_figures=True,
                                           user_naming_rule='target')

# %%
# Plot target and rescued signal
# ---------------------------------------------------------------------------

pipeline['plotter_4'] = PlotWaveforms(pipeline['fft_ave_no_artefacts'],
                                      overlay=[pipeline['fft_target']],
                                      ch_to_plot=np.array(['O2', 'T8', 'T7']),
                                      plot_x_lim=[50, 200],
                                      plot_y_lim=[-1, 0.3],
                                      return_figures=True,
                                      user_naming_rule='waves_target_rescued')

pipeline.run()

# %%
# Get generated data and save to database for data with and without dss
# ---------------------------------------------------------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'FFR'}
database_path = reader.input_node.paths.file_directory + os.sep + 'ffr_ht2_test.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave_no_artefacts'],
                                                      pipeline['fft_ave_with_artefacts']],
                                      include_waveforms=False
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
import peegy.processing.tools.weightedAverage as averager
import numpy as np
import matplotlib.pyplot as plt
import pyqtgraph as pg
# import pathos.multiprocessing as mp
fs = 56000.0
time = np.array(list(range(2**12))) * 1/fs

f = [1000.0, 109.0, 500.0]
cycles = 1.5 / f[0] * fs
# signal = -np.sin(2 * np.pi * f[0] * time / fs - 0 * np.pi / 2) * np.exp(-time / (1 / f[0]))

signal = -np.sin(2 * np.pi * f[0] * time)


signal = signal - np.mean(signal)
std_signal = np.std(signal)
N = 1000
desired_snr = 25.0
ini_std = 10.0 ** (-desired_snr / 20.0) * std_signal * N ** 0.5
theoretical_rn = ini_std / N ** 0.5

j_ave = averager.JAverager()
j_ave.splits = 1
j_ave.fs = fs
j_ave.t_p_snr = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]) * 1e-3
j_ave.analysis_window = np.array([]) * 1e-3
j_ave.time_offset = 0.0
j_ave.alpha_level = 0.05
j_ave.min_block_size = 32
j_ave.frequencies_to_analyze = np.array(f)
j_ave.plot_sweeps = False
j_ave.rejection_level = np.Inf
j_ave.rejection_window = np.array([0, 2])
j_ave.plot_frequency_range = np.array([0, 1200])
j_ave.low_pass = 0
j_ave.high_pass = 0


# def add_sweep(_in_q):
#     while True:
#         c = 0
#         _buffer = _in_q.get()
#         # next line solve class conflict between pint and incoming data buffer
#         j_ave.add_sweep(_buffer[0], split_id=_buffer[1])
#         c += 1
#         print("received", c)
#         print(j_ave.split_sweep_count)
#
# q = multiprocessing.Queue(maxsize=10)

# t = []
# for i in range(1):
#     t.append(Thread(target=add_sweep, args=(q,)))
#     t[-1].daemon = False
#
#     t[-1].start()

off_channels = np.arange(0, j_ave.splits, 2)
for i in range(N):
    split_id = np.mod(i, j_ave.splits)
    noise = ini_std * np.random.randn(signal.size)
    # sig_noise = noise * (np.where(off_channels == split_id)[0].size == 0) + signal
    sig_noise = noise + signal
    # q.put([sig_noise, split_id], block=True)
    j_ave.add_sweep(sig_noise, split_id=split_id)
    if not np.mod(i, 50):
        j_ave.plot_current_data()
        pg.QtCore.QCoreApplication.processEvents()
    print("sent", i)

print(["Theoretic SNR", desired_snr, "Estimated Standard SNR", 10*np.log10(j_ave.s_snr), "Estimated Weighted",
       10*np.log10(j_ave.w_snr),
       10 * np.log10(j_ave.s_snr_all_splits)])
print(["Theoretic RN", theoretical_rn, "Estimated RNs", j_ave.s_rn, "Estimated RNw", j_ave.w_rn])
plt.plot(j_ave.time_vector, np.hstack(((np.array([signal])).T, j_ave.s_average, j_ave.w_average)))
ht = j_ave.hotelling_t_square_test()
print(ht)
# j_ave.plot_noise()
# plt.show()
# plt.plot(np.transpose(np.vstack((signal,j_ave.w_average_all_splits()))))
lines = plt.plot(j_ave._fft_frequencies, np.vstack((np.abs(np.fft.rfft(signal)),
                                                    j_ave.s_fft_spectral_ave_all_splits,
                                                    j_ave.w_fft_spectral_ave_all_splits,
                                                    j_ave.w_fft_spectral_weighted_ave_all_splits(0))).T)

# l_style = ['*', '-', '--', ':']
# [plt.setp(lines[i], linestyle=l_style[i]) for i in range(len(lines))]

plt.legend(['signal', 'standard', 'weighted entire spectrum', 'wighted at bin'])
plt.figure()
plt.plot(j_ave.time_vector, np.vstack((signal, j_ave.s_average_all_splits, j_ave.w_average_all_splits,
                                       j_ave.w_average_spectral_weighted_ave_all_splits(0))).T)
plt.legend(['signal', 'standard', 'weighted entire spectrum', 'wighted at bin'])
plt.show()
plt.plot(j_ave.time_vector, j_ave.s_average)
plt.show()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import epochs_processing_tools as ept
from peegy.processing.tools.eeg_epoch_operators import w_mean
import numpy as np
import matplotlib.pyplot as plt
# create synthetic data
fs = 1000.0
nsamples = np.round(1 * fs).astype(int)
nchans = 36
ntrials = 130
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs - np.pi / 4), axis=1)
coeff = np.ones(nchans//2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)

s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))

desired_snr = 1
ini_std = 10.0 ** (-desired_snr / 20.0) * s_std * ntrials ** 0.5
theoretical_rn = ini_std / ntrials ** 0.5

noise = np.random.normal(0, ini_std[0], size=(nsamples, nchans, ntrials))
data = noise + s * 1
w, y_fft, rn, snr = ept.get_discrete_frequency_weights(
    epochs=data,
    block_size=10,
    fs=fs,
    frequency=f1,
    weighted_average=True
)

time_wa = w_mean(data, weights=w)
freq_wa = np.abs(np.fft.rfft(time_wa, axis=0)) / time_wa.shape[0]
freq_source = np.abs(np.fft.rfft(np.mean(s, axis=2), axis=0)) / s.shape[0]
diff = (y_fft - np.atleast_3d(w_mean(y_fft, weights=w[0, :, :])))
noise = np.sqrt(np.abs(np.mean(diff ** 2.0, axis=2))) / (time_wa.shape[0] * np.sqrt(diff.shape[2]))
freq = np.arange(0, freq_wa.shape[0]) * fs / time_wa.shape[0]
fig = plt.figure()
ax = fig.add_subplot(131)
ax.plot(freq, freq_wa)
# ax.plot(freq, freq_source)
ax.plot(f1, np.abs(w_mean(y_fft, weights=w[0, :, :])) / time_wa.shape[0], 'o')
[ax.axhline(y=_n.value) for _n in rn.squeeze() / time_wa.shape[0]]
plt.show()



"""
.. _tut-acc-eog-removal-sim:

##################################
EOG removal (Simulated)
##################################

This example illustrates removal of the eye artefacts via a template subtraction method
described in Valderrama et al, 2018. This script will output a figure, showing the
estimated eye artefact, and the EEG signal before and after eye artefact removal.
The algorithm performs three steps:
1) Blinks are detected in each channel using a (generic) eye artefact template.
The template is iteratively adjusted after each newly detected eye blink, to more closely match
the individual subject one.
2) A signal is generated, which used the individual template to represent all detected eyeblinks,
depending on the amplitudes and timepoints of each individual eyeblink.
3) The signal from step two is subtracted from the raw EEG signal, thus leading to an enhanced
EEG signal, where eyeblinks do not lead to dropping of the affected epochs.
Literature:
Valderrama, J. T., de la Torre, A., & Van Dun, B. (2018). An automatic algorithm for blink-artifact suppression based
on iterative template matching: Application to single channel recording of cortical auditory evoked potentials.
Journal of Neural Engineering, 15(1), 016008. https://doi.org/10.1088/1741-2552/aa8d95

.. contents:: Page contents
   :local:
   :depth: 2
"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.general import ReferenceData, FilterData, RegressOutEOG, ReSampling, AutoRemoveBadChannels, \
    BaselineCorrection
from peegy.processing.pipe.epochs import AverageEpochs, EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np

# %%
# Time windows
# ======================
# Time windows are used to define where to look for different peaks


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=100 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=50 * u.ms, end_time=150 * u.ms, label='gfp1', positive_peak=True,
                              target_channels=['GFP']),
               TimePeakWindow(ini_time=150 * u.ms, end_time=500 * u.ms, label='gfp2', positive_peak=True,
                              target_channels=['GFP'])])

# %%
# Peak-to-peak Measures
# ============================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array([TimeROI(ini_time=100.0 * u.ms, end_time=250.0 * u.ms, measure="snr", label="itd_snr")])

# %%
# Generate some data
# =============================
# First we generate some ACC data

n_channels = 32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
event_times = np.arange(0, 100.0, 1.0)
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.1,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test_eog_template',
                           include_eog_events=True,
                           include_non_stationary_noise_events=True,
                           noise_seed=0
                           )
reader.run()

# %%
# Start the pipeline
# =============================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader, reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['down_sampled'] = ReSampling(pipeline['channel_cleaned'],
                                      new_sampling_rate=256. * u.Hz)
pipeline.run()
pipeline['eog_removed'] = RegressOutEOG(pipeline['down_sampled'],
                                        ref_channel_labels=['EOG1'],
                                        method='template',
                                        n_iterations=10,
                                        use_initial_template=True)
pipeline.run()

# %%
# Show EOG removal Output
# ------------------------------------


pipeline['eog_removed'].plot(plot_input=True,
                             plot_output=True,
                             ch_to_plot=['CP1', 'CP5', 'P7'],
                             interactive=False)

# %%
# Continue with the pipeline
# ------------------------------------


pipeline['time_filtered_data_eog_removed'] = FilterData(pipeline['eog_removed'],
                                                        high_pass=2.0 * u.Hz,
                                                        low_pass=30.0 * u.Hz)
pipeline['time_filtered_data_with_eog_artefacts'] = FilterData(pipeline['down_sampled'],
                                                               high_pass=2.0 * u.Hz,
                                                               low_pass=30.0 * u.Hz)
pipeline.run()

# %%
# Get Epochs
# ------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs_eog_removed'] = EpochData(pipeline['time_filtered_data_eog_removed'],
                                                event_code=1.0)
pipeline['time_epochs_with_eog_artefacts'] = EpochData(pipeline['time_filtered_data_with_eog_artefacts'],
                                                       event_code=1.0)
pipeline.run()
# %%
# Get DSS components for EOG free
# ------------------------------------
# Compute spatial filter based on EOG free epochs


pipeline['time_epochs_with_eog_removed_dss'] = CreateAndApplySpatialFilter(pipeline['time_epochs_eog_removed'],
                                                                           sf_components=np.arange(5),
                                                                           user_naming_rule='dss_no_artefacts',
                                                                           return_figures=True)
pipeline.run()

# %%
# Get DSS components for contaminated data
# -----------------------------------------
# Compute spatial filter based on EOG contaminated epochs


pipeline['time_epochs_with_eog_artefacts_and_dss'] = CreateAndApplySpatialFilter(
    pipeline['time_epochs_with_eog_artefacts'],
    sf_components=np.arange(5),
    return_figures=True,
    user_naming_rule='dss_eog_artefacts')
pipeline.run()

# %%
# Compute global field power (GFP)
# ------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data


pipeline['time_epochs_eog_removed_gfp'] = AppendGFPChannel(pipeline['time_epochs_eog_removed'])
pipeline['time_epochs_with_eog_artefacts_gfp'] = AppendGFPChannel(pipeline['time_epochs_with_eog_artefacts'])
pipeline['time_epochs_with_eog_artefacts_and_dss_gfp'] = AppendGFPChannel(
    pipeline['time_epochs_with_eog_artefacts_and_dss'])
pipeline['time_epochs_with_eog_removed_dss_gfp'] = AppendGFPChannel(pipeline['time_epochs_with_eog_removed_dss'])

pipeline.run()

# %%
# Base line correction
# ------------------------------------
# We correct the initial segment by removing the mean in the first 20 ms


pipeline['time_epochs_eog_removed_gfp_baseline_corrected'] = BaselineCorrection(
    pipeline['time_epochs_eog_removed_gfp'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)

pipeline['time_epochs_with_eog_artefacts_gfp_baseline_corrected'] = BaselineCorrection(
    pipeline['time_epochs_with_eog_artefacts'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)

pipeline['time_epochs_with_eog_artefacts_and_dss_gfp_baseline_corrected'] = BaselineCorrection(
    pipeline['time_epochs_with_eog_artefacts_and_dss_gfp'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)

pipeline['time_epochs_with_eog_removed_dss_gfp_baseline_corrected'] = BaselineCorrection(
    pipeline['time_epochs_with_eog_removed_dss_gfp'],
    ini_time=0 * u.ms,
    end_time=20 * u.ms)
pipeline.run()

# %%
# Compute HT2 statistics
# ------------------------------------
# Using the raw epochs we estimate Hotelling-T2 statistics


pipeline['ht2'] = HotellingT2Test(pipeline['time_epochs_eog_removed_gfp_baseline_corrected'],
                                  roi_windows=roi_windows)
pipeline['ht2_eog_artefacts'] = HotellingT2Test(pipeline['time_epochs_with_eog_artefacts_gfp_baseline_corrected'],
                                                roi_windows=roi_windows)
pipeline.run()
pipeline['ht2'].output_node.statistical_tests['hotelling_t2_time'][[
    "test_name", "df_1", "df_2", "f", "f_critic", "p_value"]].head()

# %%
# Compute average responses
# ------------------------------------
# We apply different types of averages and compute the mean


pipeline['time_average_eog_removed'] = AverageEpochs(
    pipeline['time_epochs_eog_removed_gfp_baseline_corrected'],
    roi_windows=roi_windows,
    weighted_average=False)

pipeline['time_average_with_eog_artefacts'] = AverageEpochs(
    pipeline['time_epochs_with_eog_artefacts_gfp_baseline_corrected'],
    roi_windows=roi_windows,
    weighted_average=False)

pipeline['time_w_average_with_eog_artefacts'] = AverageEpochs(
    pipeline['time_epochs_with_eog_artefacts_gfp_baseline_corrected'],
    roi_windows=roi_windows,
    weighted_average=True)

pipeline['time_average_with_eog_artefacts_dss'] = AverageEpochs(
    pipeline['time_epochs_with_eog_artefacts_and_dss_gfp_baseline_corrected'],
    roi_windows=roi_windows,
    weighted_average=True)

pipeline['time_average_with_eog_removed_dss'] = AverageEpochs(
    pipeline['time_epochs_with_eog_removed_dss_gfp_baseline_corrected'],
    roi_windows=roi_windows,
    weighted_average=True)
pipeline.run()

# %%
# Show an average response
# ------------------------------------
# The red traces show the input epochs whist the blue traces shows the average response


pipeline['time_average_eog_removed'].plot(plot_input=False,
                                          plot_output=True,
                                          ch_to_plot=['CP1', 'CP5', 'P7'],
                                          interactive=False)

# %%
# Detect peaks
# ------------------------------------
# Using the average data, we proceed detecting peaks


pipeline['data_with_peaks_eog_removed'] = PeakDetectionTimeDomain(pipeline['time_average_eog_removed'],
                                                                  time_peak_windows=tw,
                                                                  peak_to_peak_measures=pm)
pipeline['data_with_peaks_with_eog_artefacts'] = PeakDetectionTimeDomain(pipeline['time_average_with_eog_artefacts'],
                                                                         time_peak_windows=tw,
                                                                         peak_to_peak_measures=pm)

pipeline['data_with_peaks_with_eog_artefacts_weighted_average'] = PeakDetectionTimeDomain(
    pipeline['time_w_average_with_eog_artefacts'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

pipeline['data_with_peaks_with_eog_removed_dss'] = PeakDetectionTimeDomain(
    pipeline['time_average_with_eog_removed_dss'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

pipeline['data_with_peaks_with_eog_artefacts_and_dss'] = PeakDetectionTimeDomain(
    pipeline['time_average_with_eog_artefacts_dss'],
    time_peak_windows=tw,
    peak_to_peak_measures=pm)

pipeline.run()

# %%
# Show topographic maps
# ------------------------------------


pipeline['plotter_1'] = PlotTopographicMap(pipeline['data_with_peaks_eog_removed'],
                                           topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                           plot_x_lim=[0, 0.8],
                                           plot_y_lim=[-3, 3],
                                           user_naming_rule='eog_free')

pipeline['plotter_2'] = PlotTopographicMap(pipeline['data_with_peaks_with_eog_artefacts'],
                                           topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                           plot_x_lim=[0, 0.8],
                                           plot_y_lim=[-3, 3],
                                           user_naming_rule='with_eog')
pipeline['plotter_3'] = PlotTopographicMap(pipeline['data_with_peaks_with_eog_artefacts_weighted_average'],
                                           topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                           plot_x_lim=[0, 0.8],
                                           plot_y_lim=[-3, 3],
                                           user_naming_rule='with_eog_weighted_average')

pipeline['plotter_4'] = PlotTopographicMap(pipeline['data_with_peaks_with_eog_removed_dss'],
                                           topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                           plot_x_lim=[0, 0.8],
                                           plot_y_lim=[-3, 3],
                                           user_naming_rule='with_eog_removed_dss')

pipeline['plotter_5'] = PlotTopographicMap(pipeline['data_with_peaks_with_eog_artefacts_and_dss'],
                                           topographic_channels=np.array(['C4', 'CP2', 'GFP']),
                                           plot_x_lim=[0, 0.8],
                                           plot_y_lim=[-3, 3],
                                           user_naming_rule='with_eog_artefacts_and_dss')

pipeline['plotter_5'] = PlotWaveforms(pipeline['data_with_peaks_eog_removed'],
                                      overlay=[
                                          pipeline['data_with_peaks_with_eog_artefacts_and_dss'],
                                          pipeline['data_with_peaks_with_eog_artefacts_weighted_average'],
                                          pipeline['data_with_peaks_with_eog_removed_dss']],
                                      ch_to_plot=np.array(['C4', 'CP2', 'GFP']),
                                      return_figures=True)

pipeline.run()

# %%
# Get generated data and save to database
# ------------------------------------------------
# now we save our data to a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.file_directory + os.sep + 'acc_test_data_eog_template.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['data_with_peaks_eog_removed'],
                                                      pipeline['data_with_peaks_with_eog_artefacts_and_dss'],
                                                      pipeline['data_with_peaks_with_eog_artefacts_weighted_average'],
                                                      pipeline['ht2'],
                                                      pipeline['ht2_eog_artefacts'],
                                                      pipeline['time_epochs_with_eog_removed_dss']]
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=1200)



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools import epochs_processing_tools as ept
import matplotlib.pyplot as plt
from peegy.processing.tools.eeg_epoch_operators import et_unfold, et_fold
import numpy as np

# ICA in 3 dimensions ##################
# Synthesise data
fs = 44100.0
nsamples = int(np.round(1 * fs))
nchans = 36
ntrials = 100
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs + 3 * np.pi / 2), axis=1)
coeff = np.arange(nchans / 2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, -coeff)), 0)

# s = source * np.random.normal(0, 0.1, size=(1, nchans))
s = source * coeff
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))
SNR = 10
noise = np.random.normal(0, 0.1, size=(nsamples, nchans, ntrials))
n_source = SNR * s / np.std(s)
data = noise / np.std(noise) + n_source

unf_data = et_unfold(data)

s_components, s_unmixing, s_mixing, s_pwr, whitening_m = ept.et_ica_epochs(data=unf_data,
                                                                           tol=1e-4,
                                                                           iterations=10)

components = et_fold(s_components, data.shape[0])
scomps = np.mean(components, axis=2)
plt.plot(scomps + np.arange(0, scomps.shape[1]) * scomps.max())
plt.show()

plt.show()



__author__ = 'jundurraga-ucl'
import multiprocessing as mp
import numpy as np
# Define an output queue
output = mp.Queue()

# define a example function


def rand_string(x):
    """ Generates a random string of numbers, lower- and uppercase chars. """
    rand_str = np.random.random((1, 10000))
    print(rand_str)
    print('done:', x)


# Setup a list of processes that we want to run
processes = [mp.Process(target=rand_string, args=x) for x in range(4)]

# Run processes
for p in processes:
    p.start()

# Exit the completed processes
# for p in processes:
#     p.join()

# Get process results from the output queue
results = [output.get() for p in processes]

print(results)
